"""@package parmepy.domain.box

Classes for box domains description.
"""

from ..constants import *
from .domain import Domain
from .grid import CartesianGrid


class Box(Domain):
    """
    Periodic box representation.
    @note FP : BC as parameter may be better?
    @todo Have different BC
    """

    def __init__(self, dimension=3, length=[1.0, 1.0, 1.0], origin=[0., 0., 0.]):
        """
        Create a Periodic Box from a dimension, length and origin.
        Parameters dimensions must coincide. Raise ValueError in case of inconsistent parameters dimensions.

        By defaults, it creates a \f$[0;1]^3\f$ Box.

        @param dimension : Box dimension. Default: 3
        @param length : Box length. Default [1.0, 1.0, 1.0]
        @param origin : Box minimum position. Default [0., 0., 0.]

        >>> import parmepy as pp
        >>> import numpy as np
        >>> b = pp.Box()
        >>> (b.max == np.asarray([1.0, 1.0, 1.0])).all()
        True

        """
        if not (dimension == len(length) and dimension == len(origin)):
            raise ValueError("Box parameters inconsistents dimensions")
        Domain.__init__(self, dimension)
        ##  Box length.
        self.length = np.asarray(length, dtype=PARMES_REAL)
        ##  Box origin
        self.origin = np.asarray(origin, dtype=PARMES_REAL)
        ## Maximum Box position. max = origin + length
        self.max = self.origin + self.length
        ## Boundary conditions type
        self.boundaries = np.zeros((self.dimension), dtype=PARMES_INTEGER)
        self.boundaries[:] = PERIODIC

    def discretize(self, resolution):
        """
        Box discretization method.
        Creates a CartesianGrid from the Box and a resolution.

        @param resolution : resolution of the discretization.
        @return CartesianGrid as discretized Box.

        Use an array for discreteDomain??? i.e to associate several discretisations for the same domain.
        Or maybe this should be a member of the fields??

        >>> import parmepy as pp
        >>> import numpy as np
        >>> b = pp.Box()
        >>> n = 11
        >>> g = b.discretize(np.asarray([n, n, n]))
        >>> (g.origin == b.origin).all()
        True
        >>> (g.length == b.length).all()
        True
        >>> (g.step == b.length/(n)).all()
        True
        >>> b.discreteDomain == g
        True

        """
        self.discreteDomain = CartesianGrid(resolution, self)
        return self.discreteDomain

    def __str__(self):
        """
        Informations display.
        @return Informations
        """
        s = str(self.dimension) + "D box (parallelepipedic or rectangular) domain : \n"
        s += "   origin : " + str(self.origin) + ", maxPosition :" + str(self.max) + ", lengths :" + str(self.length) + "."
        if self.discreteDomain is not None:
            s += "Discretised on : \n" + str(self.discreteDomain)
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Box."
    print Box.__doc__
