"""
@package parmepy.domain.topology
MPI Topologies
"""
import mpi4py.MPI as MPI
from ..constants import *
import sys
import time


class CartesianTopology(object):
    """
    Define an MPI cartesian topology with an associated mesh for each sub-domain.
    Keyword arguments :
    - domain : the geometry; it must be a box.
    - resolution : global resolution
    - dim : dimension of the topology
    - comm : MPI communicator used to create this topology
    - periods : periodicity of the topology (in each direction)
    - ghosts : number of points in the ghost layer

    """
    def __init__(self, domain, resolution, dim=None, dims=None, comm=MPI.COMM_WORLD, periods=None, ghosts=None):
        ## Associated domain
        self.domain = domain
        # Compute the "optimal" processus distribution for each direction of the grid topology
        nbprocs = comm.Get_size()
        if(dims is None):
            assert(dim is not None)
            self.dims = np.asarray(MPI.Compute_dims(nbprocs, dim), dtype=PARMES_INTEGER)
            # Topology dimension
            self.dim = dim
        else:
            self.dims = np.asarray(dims, dtype=PARMES_INTEGER)
            self.dim = self.dims.size
        # Define the topology
        # default period status is periodic
        if(periods is None):
            self.periods = self.dim * (True,)
        else:
            self.periods = periods
        self.topo = comm.Create_cart(self.dims, periods=self.periods)

        # Size of the topology
        self.size = self.topo.Get_size()
        # Rank of the current process, in the new topology
        self.rank = self.topo.Get_rank()
        if (self.rank == 0) :
            print 'dims = ' , self.dims
        # Coordinates of the current process
        self.coords = self.topo.Get_coords(self.rank)
        # Neighbours
        #        self.neighbours = np.zeros((dim,2))
        self.down, self.up = self.topo.Shift(XDIR, 1)
        if(self.dim > 1):
            self.west, self.east = self.topo.Shift(YDIR, 1)
        if(self.dim > 2):
            self.south, self.north = self.topo.Shift(ZDIR, 1)
        if self.dim == 3:
            self.tabSort = np.array([[self.up, 0], [self.down, 1],
                                     [self.east, 2], [self.west, 3],
                                     [self.north, 4], [self.south, 5]])
        else:
            self.tabSort = np.array([[self.up, 0] , [self.down, 1],
                                     [self.east, 2], [self.west, 3]])
        tabSort1 = self.tabSort[self.tabSort[:,0]<=self.rank]
        tabSort2 = self.tabSort[self.tabSort[:,0]>self.rank]
        if (np.asarray(tabSort1[:,0].shape)>0):
            tabSort1 = tabSort1[tabSort1[:,0].argsort()]
        if (np.asarray(tabSort2[:,0].shape)>0):
            tabSort2 = tabSort2[tabSort2[:,0].argsort()]
            tabSort2 = tabSort2[np.invert(tabSort2[:,0].argsort() )]
        self.tabSort = np.concatenate((tabSort1, tabSort2))
        self.tabSort=self.tabSort[self.tabSort[:,0].argsort()]
        # ghost layer (default = 0)
        if(ghosts is None):
            self.ghosts = np.zeros((self.domain.dimension), dtype=PARMES_INTEGER)
        else:
            self.ghosts=np.asarray(ghosts, dtype=PARMES_INTEGER)
        # Global resolution
        self.resolution = np.asarray(resolution, dtype=PARMES_INTEGER)
        nbpoints = self.resolution[:self.dim] // self.dims
        remainingPoints = self.resolution[:self.dim] % self.dims
        localResolution = self.resolution.copy()
        localResolution[:self.dim] = nbpoints
        for i in range(self.dim):
            if(self.coords[i] == self.dims[i] - 1):
                localResolution[i] = nbpoints[i] + remainingPoints[i]
        start = np.zeros((domain.dimension), dtype=PARMES_INTEGER)
        start[:self.dim] = self.coords * nbpoints
        localResolution = localResolution - 1 * self.periods
        self.mesh = LocalMesh(self.rank,
                              resolution=localResolution,
                              start=start,
                              dom_size=self.domain.length / (self.resolution-1),
                              dom_origin=self.domain.origin,
                              ghosts=self.ghosts)

    def __str__(self):
        """ Topology info display """
        s = '=======================================================\n'
        if(self.rank == 0):
            s += str(self.dim) + 'D cartesian Topology of size ' + str(self.dims) + '\n'
        s += str(self.mesh)
        s += '=======================================================\n'
        return s


class FFTTopology(object):
    """
    A 2D or 3D MPI cartesian topology
    """

    def __init__(self, dim=2, comm=MPI.COMM_WORLD, nbprocs=MPI.COMM_WORLD.Get_size(), periods=3 * (True,), dims=(1, 1, 1)):
        # Topology dimension
        self.dim = dim
        # Compute the "optimal" processus distribution for each direction of the grid topology
        dims = MPI.Compute_dims(nbprocs, dim)
        self.dims = np.ones((dim + 1), dtype=PARMES_INTEGER)
        self.dims[0:dim] = np.asarray(dims, dtype=PARMES_INTEGER)
        # Define the topology
        self.topo = comm.Create_cart(self.dims, periods=periods[0:dim + 1])
        # Size of the topology
        sizeT = self.topo.Get_size()
        # Rank of the current process
        self.rank = self.topo.Get_rank()
        # Coordinates of the current process
        self.coords = self.topo.Get_coords(self.rank)
        # Neighbours
        # self.neighbours = np.zeros((dim,2))
        self.down, self.up = self.topo.Shift(XDIR, 1)
        if(dim > 1):
            self.west, self.east = self.topo.Shift(YDIR, 1)
        if(dim > 2):
            self.south, self.north = self.topo.Shift(ZDIR, 1)

    def __str__(self):
        """ Topology info display """
        s = '======================================================='
        s += str(self.dim) + 'D FFT Topology of size ' + str(self.dims)
        s += self.mesh.__str__
        s += '======================================================='
        return s


class LocalMesh(object):
    """
    """
    def __init__(self, rank, resolution, start, dom_size, dom_origin, ghosts=np.zeros((3))):
        # Current process rank in the topology that owns this mesh
        self.rank = rank
        # Local resolution
        self.resolution = resolution.copy() + 2 * ghosts
        # index of the lower point (in each dir) in the global mesh
        self.start = start  # - ghosts
        # index of the upper point (in each dir), global mesh
        self.end = self.start + self.resolution - 1
        # Mesh step size
        self.size = dom_size
        # Mesh coordinates
        coord_start = self.start * self.size + dom_origin - ghosts* self.size
        self.origin = coord_start
        coord_end = (self.end +1) * self.size + dom_origin - ghosts* self.size
#        print 'Print TOPO'
#        print 'resolution', self.resolution , 'shape', (coord_end-coord_start)/self.size, 'coord_end',coord_end, 'coord_start',coord_start
        if self.start.size == 3:
            self.coords = np.ogrid[coord_start[0]:coord_end[0]:self.size[0],
                                   coord_start[1]:coord_end[1]:self.size[1],
                                   coord_start[2]:coord_end[2]:self.size[2]]
        elif self.start.size == 2:
            self.coords = np.ogrid[coord_start[0]:coord_end[0]:self.size[0],
                                   coord_start[1]:coord_end[1]:self.size[1]]
        else:
            self.coords = np.ogrid[coord_start[0]:coord_end[0]:self.size[0]]

    def __str__(self):
        """ Local mesh display """
        s = '[' + str(self.rank) + '] Local mesh resolution :' + str(self.resolution) + '\n'
        s += 'Global indices :' + str(self.start) + '\n'
        return s


if __name__ == "__main__":
    print "This module defines the following classes:"
    print "- CartesianTopology: ", CartesianTopology.__doc__
    print "- FFTTopology: ", FFTTopology.__doc__
    print "- LocalMesh: ", LocalMesh.__doc__
