!> Interface to scales-advection solver
module scales2py

use cart_topology, only : cart_create,set_group_size,discretisation_create,N_proc,coord,cart_rank
use advec, only : advec_init,advec_step
use mpi
use parmesparam

implicit none

contains

  !> Initialisation of advection solver (from scales) context : create topology and memory buffers
  !! @param[in] ncells number of cells in the global discrete domain
  !! @param[in] lengths width of each side of the domain
  !! @param[in] topodims number of mpi-processus in each dir
  !! @param[out] datashape local dimension of the input/output field
  !! @param[out] offset absolute index of the first component of the local field
  subroutine init_advection_solver(ncells,lengths,topodims,datashape,offset,dim,order,stab_coeff)
    integer, intent(in) :: dim 
    integer, dimension(dim),intent(in) :: ncells
    real(pk),dimension(dim), intent(in) :: lengths
    integer, dimension(dim), intent(in) :: topodims
    integer(ik), dimension(dim), intent(out) :: datashape
    integer(ik), dimension(dim), intent(out) :: offset
    character(len=*), optional ::  order
    real(pk), optional, intent(out) ::  stab_coeff
    !f2py optional , depend(ncells) :: dim=len(ncells)
    !f2py intent(hide) dim
    !f2py character(*) optional, intent(in) :: order = 'p_O2'
    !f2py, depends(dim), intent(in) :: topodims
    integer :: error,groupsize !rank, nbprocs

    if(dim /= 3) then
       stop 'Scales advection solver initialisation failed : not yet implemented for 2d problem.'
    end if

    ! get current process rank
    !call MPI_COMM_RANK(MPI_COMM_WORLD,rank,error)
    !call MPI_COMM_SIZE(MPI_COMM_WORLD,nbprocs,error)

    !groupsize = 5
    call cart_create(topodims,error)
    !call set_group_size(groupSize)
    ! Create meshes
    call discretisation_create(ncells(1),ncells(2),ncells(3),lengths(1),lengths(2),lengths(3))
    
    ! Init advection solver
    call advec_init(order,stab_coeff)
    
    ! get the local resolution (saved in scales global variable "N_proc")
    datashape = N_proc + 1
    ! get offset (i.e. global index of the lowest point of the current subdomain == scales global var "coord")
    offset = coord
    
  end subroutine init_advection_solver

  !> Particular solver for the advection of a scalar 
  !! @param[in] dt current time step
  !! @param[in] vx x component of the velocity used for advection
  !! @param[in] vy y component of the velocity used for advection
  !! @param[in] vz z component of the velocity used for advection
  !! @param[in,out] scal 3d scalar field which is advected
  subroutine solve_advection(dt,vx,vy,vz,scal)

    real(pk), intent(in) :: dt
    real(pk), dimension(:,:,:), intent(in) :: vx, vy, vz
    real(pk), dimension(size(vx,1),size(vx,2),size(vx,3)), intent(inout) :: scal
    !f2py intent(in,out), depend(size(vx,1)) :: scal

    real(pk) :: t0

    t0 = MPI_Wtime()
    call advec_step(dt,vx,vy,vz,scal)
    print *, "inside ...", cart_rank, ":", MPI_Wtime()-t0
  end subroutine solve_advection

end module scales2py
