"""
@package parmepy.fields.discrete

Discretes variables (scalar and vectors) descriptions.
"""
import numpy as np
from ..constants import *
#import evtk.hl as evtk


class ScalarField(object):
    """
    A scalar field, defined on each grid of each subdomain of the given topology.

    """

    def __init__(self, parent, topology=None, name="?", idFromParent=None):
        """
        Creates a ScalaField.

        @param parent : Continuous field.
        @param topology : Topology informations
        @param name : Field name
        @param idFromParent : Index in the parent's discrete fields
        """
        if(topology is not None):
            ## Topology informations.
            self.topology = topology
        else:
            raise NotImplementedError()
        ## Field name.
        self.name = name
        ## Field dimension.
        self.dimension = topology.domain.dimension
        ## Field resolution.
        self.resolution = self.topology.mesh.resolution
        ## Field data numpy array. \todo is numpy array zeros or empty???
        self.data = np.zeros((self.resolution), dtype=PARMES_REAL, order=ORDER)
        ## Pointer to gpu data array if needed.
        self.gpu_data = None
        ## @private Continuous field
        self.__parentField = parent
        ## @private Index in parent's field discrete fields
        self.__idFromParent = idFromParent
        ## Application domain of the variable.
        self.domain = self.__parentField.domain.discreteDomain
        ## Is self.data contains data
        self.contains_data = False
        ## Scalar is not vector
        self.vector = False

    def __str__(self):
        """ Display class information. """
        s = '[' + str(self.topology.rank) + '] '
        s += " id " + str(self.__idFromParent) + ", "
        s += self.name + " " + str(self.dimension) + 'D discrete field with resolution '
        s += str(self.data.shape) + "."
        return s + "\n"

    def __getitem__(self, i):
        """
        Access to the content of the field.
        @param i : requested index.
        @return data[i].
        """
        return self.data.__getitem__(i)

    def __setitem__(self, i, value):
        """
        Set the content of the field component at position i.

        @param i : requested index.
        @param value : value to set.

        Usage :\n
        A = ScalarField(topo)\n
        A[2,1,1] = 12.0 # Calls A.data[2,1,1] = 12.0
        """
        self.data.__setitem__(i, value)

    def initialize(self, formula=None):
        """
        Initialize values with given formula.

        @param formula : formula to apply on coordinates.

        Formula need to be vectorized by numpy. Therefore formula must have the following signature:
        @li 3D : float formula(float x, float y, float z)
        @li 2D : float formula(float x, float y)
        where x,y,z are point coordinates.
        """
        if formula is not None:
            print "...",
            v_formula = np.vectorize(formula)
            if self.dimension == 3:
                self.data = v_formula(self.topology.mesh.coords[0][0:self.resolution[0],0:self.resolution[1],0:self.resolution[2]],
                                      self.topology.mesh.coords[1][0:self.resolution[0],0:self.resolution[1],0:self.resolution[2]],
                                      self.topology.mesh.coords[2][0:self.resolution[0],0:self.resolution[1],0:self.resolution[2]])
            elif self.dimension == 2:
                self.data = v_formula(self.topology.mesh.coords[0][0:self.resolution[0],0:self.resolution[1]],
                                      self.topology.mesh.coords[1][0:self.resolution[0],0:self.resolution[1]])
            else:
                self.data = v_formula(self.topology.mesh.coords[0][0:self.resolution[0]])
            self.contains_data = True
        else:
            print "No formula",


class VectorField(object):
    """
    A vector field, defined on each grid of each subdomain of the given topology.
    Vector field is composed of several ScalarFields
    """

    def __init__(self, parent, topology=None, name="?", idFromParent=None):
        """
        Creates a VectorField.

        @param parent : Continuous field.
        @param topology : Topology informations
        @param name : Field name
        @param idFromParent : Index in the parent's discrete fields
        """
        if(topology is not None):
            ## Topology informations.
            self.topology = topology
        else:
            raise NotImplementedError()
        ## Field name.
        self.name = name
        ## Field dimension.
        self.dimension = topology.domain.dimension
        ## Field resolution.
        self.resolution = self.topology.mesh.resolution
        ## Field data numpy array. \todo is numpy array zeros or empty???
        self.data = [np.zeros((self.resolution), dtype=PARMES_REAL, order=ORDER) for d in xrange(self.dimension)]
        ## Pointer to gpu data arrays if needed.
        self.gpu_data = [None for d in xrange(self.dimension)]
        ## @private Continuous field
        self.__parentField = parent
        ## @private Index in parent's field discrete fields
        self.__idFromParent = idFromParent
        ## Application domain of the variable.
        self.domain = self.__parentField.domain.discreteDomain
        ## Is self.data contains data
        self.contains_data = False
        ## Scalar is not vector
        self.vector = True

    def __str__(self):
        """ Display class information """
        s = '[' + str(self.topology.rank) + '] '
        s += " id " + str(self.__idFromParent) + ", "
        s += self.name + " " + str(self.dimension) + 'D discrete vector field with resolution '
        s += str([data.shape for data in self.data]) + "."
        return s + "\n"

    def __getitem__(self, i):
        """ Access to the content of the field.

        @param i : requested index.
        @return data[i] depending on i.

        Usage (3D): \n
        A = VectorField(...), We have 3 components len(a.data) == 3. \n
        Following instructions access to index 2,1,1 of y component:
        @li A[1,2,1,1]
        @li A[1][2,1,1]
        @li Access to whole vector of index 2,1,1: A[2,1,1]
        @note Access to all datas as an array : A[:,:,:] (resulting shape is (dim, Nx, Ny, Nz))
        @note Access to all datas in y component as an array : A[1][:,:,:] (resulting shape is (Nx, Ny, Nz))
        """
        try:
            if len(i) == len(self.data):
                return np.asarray([data.__getitem__(i) for data in self.data])
            else:
                return self.data[i[0]].__getitem__(tuple(i[1:]))
        except (TypeError):
            return self.data[i]

    def __setitem__(self, i, value):
        """
        Set the content of the vector field component at position i.

        @param i : requested index.
        @param value : value to set.

        Usage :\n
        A = VectorField(topo)\n
        A[2,1,1] = 12.0 # Calls A.data[d][2,1,1] = 12.0 for all components d.\n
        A[2,1,1] = [12.0, 13.0, 14.0] # Calls A.data[0][2,1,1] = 12.0, A.data[1][2,1,1] = 13.0 and A.data[2][2,1,1] = 14.0\n
        A[1][2,1,1] = 13.0 # Calls A.data[1][2,1,1] = 12.0
        A[1,2,1,1] = 13.0 # Calls A.data[1][2,1,1] = 12.0
        """
        if len(i) == len(self.data):
            try:
                [data.__setitem__(i, v) for data, v in zip(self.data, value)]
            except (TypeError):
                [data.__setitem__(i, value) for data in self.data]
        elif len(i) > len(self.data):
            self.data[i[0]].__setitem__(tuple(i[1:]), value)

    def initialize(self, formula=None):
        """
        Initialize values with given formula.

        @param formula : formula to apply on coordinates.

        Formula need to be vectorized by numpy. Therefore formula must have the following signature:
        @li 3D : float, float, float formula(float x, float y, float z)
        @li 2D : float, float formula(float x, float y)
        where x,y,z are point coordinates.
        """
        if formula is not None:
            print "...",
            v_formula = np.vectorize(formula)
            if self.dimension == 3:
                self.data[0], self.data[1], self.data[2] = v_formula(self.topology.mesh.coords[0][0:self.resolution[0],0:self.resolution[1],0:self.resolution[2]],
                                                                     self.topology.mesh.coords[1][0:self.resolution[0],0:self.resolution[1],0:self.resolution[2]],
                                                                     self.topology.mesh.coords[2][0:self.resolution[0],0:self.resolution[1],0:self.resolution[2]])
                self.data[0] = np.array(self.data[0], dtype=PARMES_REAL , order=ORDER)
                self.data[1] = np.array(self.data[1], dtype=PARMES_REAL , order=ORDER)
                self.data[2] = np.array(self.data[2], dtype=PARMES_REAL , order=ORDER)
            elif self.dimension == 2:
                self.data[0], self.data[1] = v_formula(self.topology.mesh.coords[0][0:self.resolution[0],0:self.resolution[1]],
                                                       self.topology.mesh.coords[1][0:self.resolution[0],0:self.resolution[1]])
            self.contains_data = True
        else:
            print "No formula",

if __name__ == "__main__":
    print __doc__
    print "- Provided class : ScalarField, VectorField"
    print ScalarField.__doc__
