"""
Discrete obstacles description
"""
import numpy as np
from ..constants import *


class Obstacle_d(object):
    """ 
    Discrete obstacles representation.
    """

    def __init__(self, parent, topology=None, name='sphere', idFromParent=None):
        """
        Creates discrete obsctacle and z-boundaries and returns 3 arrays corresponding
        to the three different porous media.
        porosity in the solid = 0
        porosity in the fluid = infinity

        @param parent : Continuous obstacle.
        @param topology : Topology informations
        @param name : Obstacle name
        @param idFromParent : Index in the parent's discrete obstacles
        """

        ## Topology of the obstacle
        if(topology is not None):
            self.topology = topology
        else:
            raise NotImplementedError()
        ## Name of the obstacle
        self.name = name
        ## Obstacle dimension
        self.dimension = topology.domain.dimension
        ## Obstacle resolution
        self.resolution = topology.mesh.resolution
        ## @private Continuous obstacle
        self.__parentObstacle = parent
        ## @private Index in parent's discrete obstacle
        self.__idFromParent = idFromParent
        ## Width of z-boundaries layer
        self.zlayer = self.__parentObstacle.zlayer
        ## Radius of the obstacle
        self.radius = self.__parentObstacle.radius
        ## Position of the obstacle center
        self.center = self.__parentObstacle.center
        ## Orientation of the obstacle
        self.orientation = self.__parentObstacle.orientation
        ## Thickness of the porous layer at the obstacle surface
        self.porousLayer = self.__parentObstacle.porousLayer
        ## Chi function (array) for z-boundaries
        self.chiBoundary = None
        ## Chi function (array) for the solid
        self.chiSolid = None
        ## Chi function (array) for the porous area
        self.chiPorous = None

    def chiFunctions(self):
        """ 
        Compute the chi functions associated to z-boundaries and solid
        """
#        boolOrientation=[np.bool(0) for i in range (6)]
#        boolOrientation[self.indOrientation(self.orientation)]=True

        ## Temporary arrays
        chiBoundaryTmp = []
        chiSolidTmp = []
        chiPorousTmp = []

        step = self.topology.mesh.size
        ghosts = self.topology.ghosts
        coord_start = self.topology.mesh.origin 
        coord_end = self.topology.mesh.end * step + self.topology.domain.origin - (ghosts * step)
        layerMin = coord_start[2] + ghosts[2] * step[2] + self.zlayer 
        layerMax = coord_end[2] - ghosts[2] * step[2] - self.zlayer
        radiusMinuslayer = self.radius- self.porousLayer
        print 'start, end, layerMin, layerMax', coord_start, coord_end, layerMin, layerMax
        for k in xrange (ghosts[2], self.resolution[2] - ghosts[2]):
            rz = coord_start[2] + step[2] * k 
            for j in xrange (ghosts[1], self.resolution[1] - ghosts[1]):
                ry = coord_start[1] + step[1] * j
                for i in xrange (ghosts[0], self.resolution[0] - ghosts[0]):
                    if (rz > layerMax or rz < layerMin):
                    # we are in the z-layer (North or South):
                         chiBoundaryTmp.append([i,j,k])
#                        chiBoundaryTmp0.append(i)
#                        chiBoundaryTmp1.append(j)
#                        chiBoundaryTmp2.append(k)
                    else :
                        rx = coord_start[0] + step[0] * i
                        dist = np.sqrt((rx - self.center[0]) **2 +
                               (ry - self.center[1]) **2 +
                               (rz - self.center[2]) **2) 
                        if (self.name=='sphere'):
                            if (radiusMinuslayer < dist and dist <= self.radius + 1E-12):
                            # we are in the porous area of the sphere:
                                chiPorousTmp.append([i,j,k])
#                                chiPorousTmp0.append(i) 
#                                chiPorousTmp1.append(j)
#                                chiPorousTmp2.append(k)
                            if (dist <= radiusMinuslayer + 1E-12):
                            # we are in the solid area of the sphere:
                                chiSolidTmp.append([i,j,k])
#                                chiSolidTmp0.append(i)
#                                chiSolidTmp1.append(j)
#                                chiSolidTmp2.append(k)
                        else :
                            if (radiusMinuslayer < dist and dist <= self.radius and rx <= self.center[0]):
                            # we are in the porous area of the hemisphere:
                                chiPorousTmp.append([i,j,k])
#                                chiPorousTmp0.append(i)
#                                chiPorousTmp1.append(j)
#                                chiPorousTmp2.append(k)
                            if (dist <= radiusMinuslayer and rx <= self.center[0]):
                            # we are in the solid area of the hemisphere:
                                chiSolidTmp.append([i,j,k])
#                                chiSolidTmp0.append(i)
#                                chiSolidTmp1.append(j)
#                                chiSolidTmp2.append(k)

        # We keep only the non-zero values in the chi arrays
        self.chiBoundary = chiBoundaryTmp #np.concatenate((np.array(chiBoundaryTmp0), np.array(chiBoundaryTmp1), np.array(chiBoundaryTmp2)))
        self.chiPorous = chiPorousTmp #np.concatenate((np.array(chiPorousTmp0), np.array(chiPorousTmp1), np.array(chiPorousTmp2)))
        self.chiSolid = chiSolidTmp #np.concatenate((np.array(chiSolidTmp0), np.array(chiSolidTmp1), np.array(chiSolidTmp2)))

    def indOrientation(orientation):
        return {'West':0, 'East':1, 'South':2, 'North':3, 'Forward':4, 'Backward':5}.get(orientation, 0)

    def __str__(self):
        """ToString method"""
        return "Obstacle_d (discrete obstacles)"

if __name__ == "__main__" :
    print "This module defines the following classes:"
    print "Obstacle_d: ", Obstacle_d.__doc__
