# -*- coding: utf-8 -*-
"""
@package operator
Discrete Diffusion operator using FFT
"""
from ..f2py import fftw2py
import mpi4py.MPI as MPI
from .discrete import DiscreteOperator
from ..constants import *
import numpy as np
import time
import sys

class Diffusion_d(DiscreteOperator):
    """
    Operator representation.
    """

    def __init__(self, diff, idVelocityD=0, idVorticityD=0, viscosity=0.1, topology=None):
        """
        Constructor.
        @param velocity : descretized velocity to use.
        @param vorticity : discretized vorticity to update.
        """
        DiscreteOperator.__init__(self)
        self.diff = diff
        self.velocity = diff.velocity.discreteField[idVelocityD]
        self.vorticity = diff.vorticity.discreteField[idVorticityD]
        self.viscosity = viscosity
        self.topology = topology
        self.ghosts = topology.ghosts
        self.resolution = topology.mesh.resolution
        self.dim = topology.dim
#        self.ppfft = parmepy.f2py.fftw2py
        self.compute_time = 0.

    def apply(self, t, dt):
        """
        Apply operator.
        """
        self.compute_time = time.time()

        ind0a = self.ghosts[0]
        ind0b=self.resolution[0]-self.ghosts[0]
        ind1a = self.ghosts[1]
        ind1b=self.resolution[1]-self.ghosts[1]
        ind2a = self.ghosts[2]
        ind2b=self.resolution[2]-self.ghosts[2]
        ind0=np.arange(ind0a,ind0b)
        ind1=np.arange(ind1a,ind1b)
        ind2=np.arange(ind2a,ind2b)

        vorticityNoG = [np.zeros((self.resolution - 2 * self.ghosts), dtype=PARMES_REAL, order=ORDER) for d in xrange(self.dim)]
        velocityNoG = [np.zeros((self.resolution - 2 * self.ghosts), dtype=PARMES_REAL, order=ORDER) for d in xrange(self.dim)]
        for i in xrange (self.dim):
            vorticityNoG[i] = self.vorticity[i][ind0a:ind0b,ind1a:ind1b,ind2a:ind2b]
            velocityNoG[i] = self.velocity[i][ind0a:ind0b,ind1a:ind1b,ind2a:ind2b]

        # Vorticity diffusion
        vorticityNoG[0][...], vorticityNoG[1][...], vorticityNoG[2][...] = \
            fftw2py.solve_diffusion_3d(self.viscosity * dt, 
                                          velocityNoG[0][...], velocityNoG[1][...], velocityNoG[2][...], 
                                          vorticityNoG[0][...], vorticityNoG[1][...], vorticityNoG[2][...])

        for i in xrange (self.dim):
            self.vorticity[i][ind0a:ind0b,ind1a:ind1b,ind2a:ind2b] = vorticityNoG[i][...]

#        print 'shape vorti, velo, diffusion', np.asarray(self.vorticity.data).shape, np.asarray(self.velocity.data).shape
        self.compute_time = time.time() - self.compute_time
        self.total_time += self.compute_time
        return self.compute_time
#        return result

    def printComputeTime(self):
        self.timings_info[0] = "\"Diffusion calculation total\""
        self.timings_info[1] = str(self.total_time)
#        print "Diffusion total time : ", self.total_time
        print "Time of the last diffusion calculation loop :", self.compute_time 

    def __str__(self):
        s = "Diffusion_d (DiscreteOperator). " + DiscreteOperator.__str__(self)
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Diffusion_d"
    print Diffusion_d.__doc__
