"""
@package parmepy.operator.discrete

Discrete operator representation.
"""
from abc import ABCMeta, abstractmethod


class DiscreteOperator:
    """
    Abstract description of a discretized operator.
    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self):
        """
        Create an empty discrete operator.
        """
        ## Input variables
        self.input = None
        ## Output variables
        self.output = None
        ## variables
        self.variables = []
        ## Operator numerical method.
        self.numMethod = None
        ## DiscreteOperator is a discrete operator
        self.discreteOperator = self
        ## Total compute time
        self.total_time = 0.
        ## Operator name
        self.name = "?"
        ## Timings informations
        self.timings_info = ["", ""]

    def setMethod(self, method):
        """
        Set the numerical method used by operator when calling applyOperator.

        @param method : Numerical method.
        """
        self.numMethod = method

    def addVariable(self, cVariable):
        """
        Add an continuous variables to the operator.

        @param cVariable : list of variables to add.
        """
        for v in cVariable:
            if self.variables.count(v) == 0:
                self.variables.append(v)

    @abstractmethod
    def printComputeTime(self):
        """Print total computing time."""
        raise NotImplementedError("Need to override method in a subclass of DiscreteOperator")

    @abstractmethod
    def apply(self):
        """
        Abstract method, apply operaton on a variable.
        Must be implemented by sub-class.
        """
        raise NotImplementedError("Need to override method in a subclass of DiscreteOperator")

    def __str__(self):
        """ToString method"""
        s = "Applying on following fields : \n"
        if self.input is not None:
            s += "Input fields : \n"
            for f in self.input:
                s += str(f) + "\n"
        if self.output is not None:
            s += "Output fields : \n"
            for f in self.output:
                s += str(f) + "\n"
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : DiscreteOperator"
    print DiscreteOperator.__doc__
