# -*- coding: utf-8 -*-
"""
@package operator
Convert differential operator for stretching into a function to Discrete time integration
"""
from ..constants import *
from .discrete import DiscreteOperator
from .differentialOperator_d import DifferentialOperator_d
import numpy as np

class Fct2Op(object):
    """
    Fct2Op 
    
    make the link between differentialOperator_d and the integration time scheme (Euler,RK,...)
    """

    def __init__(self, field1, field2, choice=None, topology=None):
        """
        Constructor.
        store field2 to work with it in the apply function

        @param field1 : vorticity field 
        @param field2 : velocity field or grad(U) depending on 'choice' parameter
        @param choice : can be 'gradV' for w.Du or divWU for div(wu)
        @param resolution : resolution of the integration domain (array of number of points)
        @param meshSize : array of the space step
        """
        self.field1 = field1
        self.field2 = field2
        self.choice = choice
        self.topology = topology
        self.resolution = self.topology.mesh.resolution
        self.meshSize = self.topology.mesh.size

    def apply(self, t , y):
        """
        Apply operator.
        
        We used Fct2Op.apply like a def function
        """
        if self.choice == 'gradV' : # field2 = gradU
            temp0 = y[0][...] * 0.
            temp1 = y[0][...] * 0.
            temp2 = y[0][...] * 0.
            ind0a = self.topology.ghosts[0]
            ind0b=self.resolution[0]-self.topology.ghosts[0]
            ind1a = self.topology.ghosts[1]
            ind1b=self.resolution[1]-self.topology.ghosts[1]
            ind2a = self.topology.ghosts[2]
            ind2b=self.resolution[2]-self.topology.ghosts[2]
            ind0=np.arange(ind0a,ind0b)
            ind1=np.arange(ind1a,ind1b)
            ind2=np.arange(ind2a,ind2b)
            #Make Scalar product
            temp0[ind0a:ind0b,ind1a:ind1b,ind2a:ind2b] = self.field2[0][ind0a:ind0b,ind1a:ind1b,ind2a:ind2b]*\
                                                            y[0][ind0a:ind0b,ind1a:ind1b,ind2a:ind2b] + \
                                                        self.field2[1][ind0a:ind0b,ind1a:ind1b,ind2a:ind2b]* \
                                                            y[1][ind0a:ind0b,ind1a:ind1b,ind2a:ind2b] + \
                                                        self.field2[2][ind0a:ind0b,ind1a:ind1b,ind2a:ind2b]* \
                                                            y[2][ind0a:ind0b,ind1a:ind1b,ind2a:ind2b]
            temp1[ind0a:ind0b,ind1a:ind1b,ind2a:ind2b] = self.field2[3][ind0a:ind0b,ind1a:ind1b,ind2a:ind2b]* \
                                                            y[0][ind0a:ind0b,ind1a:ind1b,ind2a:ind2b] + \
                                                        self.field2[4][ind0a:ind0b,ind1a:ind1b,ind2a:ind2b]* \
                                                            y[1][ind0a:ind0b,ind1a:ind1b,ind2a:ind2b] + \
                                                        self.field2[5][ind0a:ind0b,ind1a:ind1b,ind2a:ind2b]* \
                                                            y[2][ind0a:ind0b,ind1a:ind1b,ind2a:ind2b]
            temp2[ind0a:ind0b,ind1a:ind1b,ind2a:ind2b] = self.field2[6][ind0a:ind0b,ind1a:ind1b,ind2a:ind2b]* \
                                                            y[0][ind0a:ind0b,ind1a:ind1b,ind2a:ind2b] + \
                                                        self.field2[7][ind0a:ind0b,ind1a:ind1b,ind2a:ind2b]* \
                                                            y[1][ind0a:ind0b,ind1a:ind1b,ind2a:ind2b] + \
                                                        self.field2[8][ind0a:ind0b,ind1a:ind1b,ind2a:ind2b]* \
                                                            y[2][ind0a:ind0b,ind1a:ind1b,ind2a:ind2b]
            result = np.concatenate((np.array([temp0]),np.array([temp1]),np.array([temp2])))
            return result
        if self.choice == 'divWU' : # field2 = velocity field
            print 'taille field22:', self.field2.shape
            result = DifferentialOperator_d(y, self.field2, self.choice, self.topology).apply()
            return result
        if self.choice == 'hessianU' : # field2 = velocity field
            gradU, maxgersh = DifferentialOperator_d(self.field1, self.field2, choice='gradV', topology=self.topology).apply()
            GradUline1 = np.concatenate(([gradU[0]],[gradU[1]],[gradU[2]]))
            GradUline2 = np.concatenate(([gradU[3]],[gradU[4]],[gradU[5]]))
            GradUline3 = np.concatenate(([gradU[6]],[gradU[7]],[gradU[8]]))
            result1, maxgersh = DifferentialOperator_d(self.field1, GradUline1, choice='gradV', topology=self.topology).apply()
            result2, maxgersh = DifferentialOperator_d(self.field1, GradUline2, choice='gradV', topology=self.topology).apply()
            result3, maxgersh = DifferentialOperator_d(self.field1, GradUline3, choice='gradV', topology=self.topology).apply()
            return result1, result2, result3
        else :
            print 'choice', choice , 'is not defined'


    def __str__(self):
        s = "fct2Op(DiscreteOperator). " + DiscreteOperator.__str__(self)
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Fct2Op"
    print Fct2Op.__doc__
