# -*- coding: utf-8 -*-
"""
@package operator
Discrete penalization representation
"""
from ..constants import *
from .discrete import DiscreteOperator
from .differentialOperator_d import DifferentialOperator_d
from ..physics.compute_forces import Compute_forces
import pyopencl as cl
import numpy as np
import time


class Penalization_d(DiscreteOperator):
    """
    Penalization operator representation.
    DiscreteOperator.DiscreteOperator specialization.
    """

    def __init__(self, penal, idVelocityD=0, idVorticityD=0, idObstacleD=0):
        """
        Constructor.
        Create a Penalization operator on a given continuous domain.
        Work on a given velocity and penalize it.

        @param penal : Penalization operator.
        """
        DiscreteOperator.__init__(self)
        ## Continous penalization operator
        self.penal = penal
        ## Velocity.
        self.velocity = penal.velocity.discreteField[idVelocityD]
        ## Vorticity.
        self.vorticity = penal.vorticity.discreteField[idVorticityD]
        ## Obstacle represented by characteristic function Chi
        self.obstacle = penal.obstacle.discreteObstacle[idObstacleD]
#        ## Topology
        self.topology = penal.obstacle.discreteObstacle[idObstacleD].topology
        ## input field
#        self.input = [self.velocity]
#        ## output field
#        self.output = [self.res_velocity]
        self.obstacle.chiFunctions()
        self.noca = Compute_forces(self.topology, self.obstacle, boxMin= [0.2, 0.2, 0.2], boxMax=[0.8, 0.8, 0.8])
        if (self.topology.rank == 0):
            self.f = open('./res/NocaForces.dat', 'w')

    def apply(self, t, dt):

        self.compute_time = time.time()

        #   self.obstacle.chiFunctions()
        coef = 1.0 / (1.0 + dt * self.penal.lambd[:])

        for k in self.obstacle.chiBoundary[:]:
            self.velocity[0][k[0],k[1],k[2]] = coef[2] * self.velocity[0][k[0],k[1],k[2]]
            self.velocity[1][k[0],k[1],k[2]] = coef[2] * self.velocity[1][k[0],k[1],k[2]]
            self.velocity[2][k[0],k[1],k[2]] = coef[2] * self.velocity[2][k[0],k[1],k[2]]

        for k in self.obstacle.chiPorous[:]:
            self.velocity[0][k[0],k[1],k[2]] = coef[1] * self.velocity[0][k[0],k[1],k[2]]
            self.velocity[1][k[0],k[1],k[2]] = coef[1] * self.velocity[1][k[0],k[1],k[2]]
            self.velocity[2][k[0],k[1],k[2]] = coef[1] * self.velocity[2][k[0],k[1],k[2]]

        for k in self.obstacle.chiSolid[:]:
            self.velocity[0][k[0],k[1],k[2]] = coef[2] * self.velocity[0][k[0],k[1],k[2]]
            self.velocity[1][k[0],k[1],k[2]] = coef[2] * self.velocity[1][k[0],k[1],k[2]]
            self.velocity[2][k[0],k[1],k[2]] = coef[2] * self.velocity[2][k[0],k[1],k[2]]

        self.vorticity.data = DifferentialOperator_d(self.velocity, self.vorticity, choice='curl', topology=self.topology).apply()

        # Computation of forces exerced on the obstacle (according to Noca's formula)
        Re = 200.
        nocares = self.noca.apply(t, dt, self.velocity, self.vorticity, Re)
        if (self.topology.rank == 0):
                # print time and forces values in the following order : time, cX, cY, cZ
                self.f.write("%s   %s   %s   %s\n" % (t, nocares[0], nocares[1], nocares[2]))

        self.compute_time = time.time() - self.compute_time
        self.total_time += self.compute_time

        return self.compute_time

    def printComputeTime(self):
        self.timings_info[0] = "\"Penalization total\""
        self.timings_info[1] = str(self.total_time)
        print "Penalization total time : ", self.total_time
        print "Time of the last Penalization iteration :", self.compute_time 

    def __str__(self):
        s = "Penalization_d (DiscreteOperator). " + DiscreteOperator.__str__(self)
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Penalization_d"
    print Penalization_d.__doc__
