# -*- coding: utf-8 -*-
"""
@package operator
Discrete Poisson solver operator using FFT
"""
from ..f2py import fftw2py
import mpi4py.MPI as MPI
from .discrete import DiscreteOperator
from ..constants import *
import numpy as np
import time
import sys

class Poisson_d(DiscreteOperator):
    """
    Operator representation.
    """

    def __init__(self, poisson, idVelocityD=0, idVorticityD=0, topology=None):
        """
        Constructor.
        @param velocity : descretized velocity to update from vorticity.
        @param vorticity
        """
        DiscreteOperator.__init__(self)
        self.poisson = poisson
        self.velocity = poisson.velocity.discreteField[idVelocityD]
        self.vorticity = poisson.vorticity.discreteField[idVorticityD]
        self.topology = topology
        self.ghosts = topology.ghosts
        self.resolution = topology.mesh.resolution
        self.dim = topology.dim
        self.coords = topology.mesh.coords
#        self.ppfft = parmepy.f2py.fftw2py
        self.compute_time = 0.

    def apply(self, t, dt):
        """
        Apply operator.
        """
        self.compute_time = time.time()

        ind0a = self.ghosts[0]
        ind0b=self.resolution[0]-self.ghosts[0]
        ind1a = self.ghosts[1]
        ind1b=self.resolution[1]-self.ghosts[1]
        ind2a = self.ghosts[2]
        ind2b=self.resolution[2]-self.ghosts[2]
        ind0=np.arange(ind0a,ind0b)
        ind1=np.arange(ind1a,ind1b)
        ind2=np.arange(ind2a,ind2b)

        vorticityNoG = [np.zeros((self.resolution - 2 * self.ghosts), dtype=PARMES_REAL, order=ORDER) for d in xrange(self.dim)]
        velocityNoG = [np.zeros((self.resolution - 2 * self.ghosts), dtype=PARMES_REAL, order=ORDER) for d in xrange(self.dim)]
        for i in xrange (self.dim):
            vorticityNoG[i] = self.vorticity[i][ind0a:ind0b,ind1a:ind1b,ind2a:ind2b]
            velocityNoG[i] = self.velocity[i][ind0a:ind0b,ind1a:ind1b,ind2a:ind2b]

        # Recover velocity field from vorticity field
        velocityNoG[0][...], velocityNoG[1][...], velocityNoG[2][...] = \
            fftw2py.solve_poisson_3d(vorticityNoG[0][...], vorticityNoG[1][...], vorticityNoG[2][...],
                                       velocityNoG[0][...], velocityNoG[1][...], velocityNoG[2][...])

        for i in xrange (self.dim):
            self.velocity[i][ind0a:ind0b,ind1a:ind1b,ind2a:ind2b] = velocityNoG[i][...]

#        #======== CORRECTION VEL X : u_x ==================

#        # computation of the current flow rate: int_y int_z uX dydz
#        debX = 0.
#        debX = np.sum(self.velocity[0][0,ind1a:ind1b,ind2a:ind2b]) 
#        debX = debX * self.topology.mesh.size[1] * self.topology.mesh.size[2]

#        self.velocity[0][ind0a:ind0b,ind1a:ind1b,ind2a:ind2b] = \
#            self.velocity[0][ind0a:ind0b,ind1a:ind1b,ind2a:ind2b] + 1. - debX/(4. * np.pi)**2

#        #======== CORRECTION VEL Y : u_y ==================

#        # computation of the current flow rate: int_y int_z uY dydz
#        debY = 0.
#        debY = np.sum(self.velocity[1][0,ind1a:ind1b,ind2a:ind2b]) 
#        debY = debY * self.topology.mesh.size[1] * self.topology.mesh.size[2]

#        # computation of omega_z circulation
#        omZbar = 0.
#        omZbar = np.sum(self.vorticity[2][ind0a:ind0b,ind1a:ind1b,ind2a:ind2b])
#        omZbar = omZbar * self.topology.mesh.size[1] * self.topology.mesh.size[2] / (4. * np.pi)**2

#        self.velocity[1][ind0a:ind0b,ind1a:ind1b,ind2a:ind2b] = \
#            self.velocity[1][ind0a:ind0b,ind1a:ind1b,ind2a:ind2b] \
#            + omZbar * self.coords[0][ind0a:ind0b] \
#            - omZbar * self.coords[0][ind0a] \
#            - debY/(4. * np.pi)**2

#        #======== CORRECTION VEL Z : u_z ==================

#        # computation of the current flow rate: int_y int_z uZ dydz
#        debZ = 0.
#        debZ = np.sum(self.velocity[2][0,ind1a:ind1b,ind2a:ind2b]) 
#        debZ = debZ * self.topology.mesh.size[1] * self.topology.mesh.size[2]

#        # computation of omega_y circulation
#        omYbar = 0.
#        omYbar = np.sum(self.vorticity[1][ind0a:ind0b,ind1a:ind1b,ind2a:ind2b])
#        omYbar = omYbar * self.topology.mesh.size[1] * self.topology.mesh.size[2] / (4. * np.pi)**2

#        self.velocity[2][ind0a:ind0b,ind1a:ind1b,ind2a:ind2b] = \
#            self.velocity[2][ind0a:ind0b,ind1a:ind1b,ind2a:ind2b] \
#            - omYbar * self.coords[0][ind0a:ind0b] \
#            + omYbar * self.coords[0][ind0a] \
#            - debZ/(4. * np.pi)**2

        # PENSER AU MPI_REDUCE !!


        self.compute_time = time.time() - self.compute_time
        self.total_time += self.compute_time
        return self.compute_time

    def printComputeTime(self):
        self.timings_info[0] = "\"Poisson solver total\""
        self.timings_info[1] = str(self.total_time)
#        print "Poisson total time : ", self.total_time
        print "Time of the last Poisson solver loop :", self.compute_time 

    def __str__(self):
        s = "Poisson_d (DiscreteOperator). " + DiscreteOperator.__str__(self)
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Poisson_d"
    print Poisson_d.__doc__
