"""
@package parmepy.operator.splitting

Splitting operator representation
"""
from .discrete import DiscreteOperator
import time


class Splitting(DiscreteOperator):
    """
    Splitting operator representation.

    Operator of operators applying in a Strang splitting.

    Implements a 2nd order splitting in 3D:
    @li X-dir, half time step
    @li Y-dir, half time step
    @li Z-dir, full time step
    @li Y-dir, half time step
    @li X-dir, half time step
    \n
    Implements a 2nd order splitting in 2D:
    @li X-dir, half time step
    @li Y-dir, full time step
    @li X-dir, half time step
    """

    def __init__(self, operators=[], dim=3):
        """
        Create a Splitting operator on a given list of operators and a dimension.

        @param operators : list of operators to split.
        @param dim : problem dimension.
        """
        DiscreteOperator.__init__(self)
        ## Operators to split
        self.operators = operators
        ## Splitting at 2nd order.
        self.splitting = []
        ## Half timestep in all directions
        # [self.splitting.append((i, 0.5)) for i in xrange(dim)]
        # [self.splitting.append((dim - 1 - i, 0.5)) for i in xrange(dim)]
        ## Half timestep in all directions but last
        [self.splitting.append((i, 0.5)) for i in xrange(dim - 1)]
        self.splitting.append((dim - 1, 1.))
        [self.splitting.append((dim - 2 - i, 0.5)) for i in xrange(dim - 1)]
        ## Compute time detailed per directions and per operators
        self.compute_time_details = [[0. for i in xrange(dim)] for op in self.operators]
        self.name = "splitting"

    def apply(self, t, dt):
        """
        Apply Remeshing operator.

        @param t : current time.
        @param dt : time step.
        """
        c_time = 0.
        for split in self.splitting:
            #print "direction : " + str(split[0]) + " dt*" + str(split[1])
            op_index = 0
            for op in self.operators:
                temp_time = op.apply(t, split[1] * dt, split[0])
                c_time += temp_time
                self.compute_time_details[op_index][split[0]] += temp_time
                op_index += 1
        self.total_time += c_time
        return c_time

    def printComputeTime(self):
        print "Splitting total time : ", self.total_time
        self.timings_info[0] = "\"Splitting total\" "
        self.timings_info[1] = str(self.total_time) + " "
        for op in self.operators:
            print "  ",
            op.printComputeTime()
            self.timings_info[0] += op.timings_info[0]
            self.timings_info[1] += op.timings_info[1]

    def __str__(self):
        s = "Splitting (DiscreteOperator). Splitting steps : \n"
        for split in self.splitting:
            s += "direction : " + str(split[0]) + " dt=" + str(split[1]) + "\n"
        s += "Concerned operators : \n"
        for op in self.operators:
            s += str(op)
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Splitting"
    print RemeshingDOp.__doc__
