# -*- coding: utf-8 -*-
"""
@package parmepy.operator.Stretching

Penalization operator representation
"""
from .continuous import ContinuousOperator
from .stretching_d import Stretching_d
from ..particular_solvers.integrator.runge_kutta3 import RK3


class Stretching(ContinuousOperator):
    """
    Stretching operator representation
    """

    def __init__(self, velocity, vorticity):
        """
        Constructor.
        Create a Stretching operator from given velocity and vorticity variables.

        @param velocity ContinuousVectorField : velocity variable.
        @param vorticity ContinuousVectorField : vorticity variable.
        """
        ContinuousOperator.__init__(self)
        ## velocity variable (vector)
        self.velocity = velocity
        ## vorticity variable (vector)
        self.vorticity = vorticity
        self.addVariable([velocity, vorticity])

    def discretize(self, topology=None, idVelocityD=0, idVorticityD=0, propertyOp='divConservation', method=RK3):
        """
        Stretching operator discretization method.
        Create a discrete Stretching operator from given specifications.

        @param idVelocityD : Index of velocity discretisation to use.
        @param idVorticityD : Index of vorticity discretisation to use.
        @param propertyOp : choice of the property garantied by the stretching Op ( divConservation or massConservation)
        @param method : the method to use. (Euler, RK, ..) default : RK3
        """
        if self.discreteOperator is None:
            self.discreteOperator = Stretching_d(self, topology, idVelocityD, idVorticityD, propertyOp, method)

    def __str__(self):
        """ToString method"""
        s = " Stretching operator (ContinuousOperator)"
        if self.discreteOperator is not None:
            s += "with the following discretization:\n"
            s += str(self.discreteOperator)
        else:
            s += ". Not discretised"
        return s + "\n"

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Stretching"
    print Stretching.__doc__
