# -*- coding: utf-8 -*-
"""
@package operator
Discrete stretching representation
"""
from ..constants import *
from .discrete import DiscreteOperator
from .differentialOperator import DifferentialOperator
from .differentialOperator_d import DifferentialOperator_d
from .fct2op import Fct2Op
from ..particular_solvers.integrator.euler import Euler
from ..particular_solvers.integrator.runge_kutta2 import RK2
from ..particular_solvers.integrator.runge_kutta3 import RK3
from ..particular_solvers.integrator.runge_kutta4 import RK4
from ..particular_solvers.integrator.integrator import ODESolver
from ..tools.cpu_data_transfer import Synchronize
import numpy as np
import numpy.testing as npt
import time
import sys


class Stretching_d(DiscreteOperator):
    """
    Stretching operator representation.
    DiscreteOperator.DiscreteOperator specialization.
    """

    def __init__(self, stretch, topology=None, idVelocityD=0, idVorticityD=0, propertyOp='divConservation', method=None):
        """
        Constructor.
        Create a Stretching operator on a given continuous domain.
        Work on a given velocity and vorticity to compute the stretching term.

        @param stretch : Stretching operator.
        """
        DiscreteOperator.__init__(self)
        self.stretch = stretch
        ## Velocity.
        self.velocity = stretch.velocity.discreteField[idVelocityD]
        ## Vorticity.
        self.vorticity = stretch.vorticity.discreteField[idVorticityD]
#        ## input fields
#        self.input = [self.velocity, self.vorticity]
        self.propertyOp = propertyOp
        self.method = method
        ## self.name = "stretching"
        self.topology = topology
        self.OpSynchronize = Synchronize(self.topology)

    def apply(self, t, dt):
        """
        Apply Stretching operator.

        @param t : current time.
        @param dt : time step.

        Stretching algorithm:
        @li 1. discretization of the divergence operator (div wu) or w grad(u) : \n
        @li 2. Time integration. Performs a method choose to resolove the equation dw/dt = div(wu) or dw/dt = w grad(u).
         -define the function of ODEsolver as the result of the first step
         -integrate with the chosen method (RK4. RK2. euler)
        """
        self.compute_time = time.time()

        if self.method is not None:

            if self.propertyOp == 'divConservation' :

                ## Space discretisation of grad(u)
                self.stretchOp = DifferentialOperator_d(self.vorticity, self.velocity, choice='gradV', topology=self.topology)
                gradResult, maxgersh = self.stretchOp.apply()

                ## Determination of Stretching time step (stability condition)
                self.ndt = self.stabilityTest(dt, maxgersh[0], self.method)
                self.dtstretch = dt/float(self.ndt)

                print "Maxgersh", maxgersh
                print "dtStretch, ndt", self.dtstretch, self.ndt
                print "dtAdvec", 1./maxgersh[1]

                ## fct2Op
                self.fonction = Fct2Op(self.vorticity.data, gradResult,choice = 'gradV', topology=self.topology)

            elif self.propertyOp == 'massConservation' :
                ## Determination of Stretching time step (stability condition)
                self.ndt = 1
                self.dtstretch = dt

                ## fct2Op
                self.fonction = Fct2Op(self.vorticity, self.velocity,choice = 'divWU', topology=self.topology)

            else:
                raise ValueError("Unknown Stretching Operator property: " + str(self.propertyOp))

            # Time integration 
            for i in range (self.ndt) :
                methodInt=self.method(self.fonction)
                self.vorticity.data = methodInt.integrate(self.fonction.apply, t, self.dtstretch, self.vorticity.data)


            # Ghosts synchronization
#            self.OpSynchronize.apply([self.vorticity.data, self.velocity.data])


            self.compute_time = time.time() - self.compute_time
            self.total_time += self.compute_time

        return self.compute_time

    def stabilityTest(self, dt, maxi, method):
        dtstretch = dt
        cststretch =0.
        if method == Euler :
            cststretch = 2.0
            dtstretch = min(dtstretch, cststretch/maxi)
        if method == RK2 :
            cststretch = 2.0
            dtstretch = min(dtstretch, cststretch/maxi)
        if method == RK3 :
            cststretch = 2.5127
            dtstretch = min(dtstretch, cststretch/maxi)
        if method == RK4 :
            cststretch = 2.7853
            dtstretch = min(dtstretch, cststretch/maxi)
        if dt == dtstretch :
            print "dt, ndt , dtstretch, upperbound", dt, int(dt/dtstretch), dt/float(int(dt/dtstretch)), cststretch/maxi
            return int(dt/dtstretch)
        else :
            print "dt, ndt , dtstretch, upperbound", dt, int(dt/dtstretch)+1, dt/float(int(dt/dtstretch)+1), cststretch/maxi
            return int(dt/dtstretch) +1

    def printComputeTime(self):
        self.timings_info[0] = "\"Stretching total\""
        self.timings_info[1] = str(self.total_time)
        print "Stretching total time : ", self.total_time
        print "Time of the last Stretching iteration :", self.compute_time

    def __str__(self):
        s = "Stretching_d (DiscreteOperator). " + DiscreteOperator.__str__(self)
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Stretching_d"
    print Stretching_d.__doc__
