"""
@package parmepy.particular_solvers.basic

Particular solver description.
"""
from solver import Solver
from .integrator.integrator import ODESolver
from .integrator.euler import Euler
from .integrator.runge_kutta2 import RK2
from .integrator.runge_kutta3 import RK3
from .integrator.runge_kutta4 import RK4
from ..fields.continuous import ContinuousField
from ..operator.transport import Transport
from ..operator.stretching import Stretching
from ..operator.remeshing import Remeshing
from ..operator.penalization import Penalization
from ..operator.advec_scales import Advec_scales
from ..operator.diffusion import Diffusion
from ..operator.poisson import Poisson
from ..operator.splitting import Splitting
from ..tools.timer import Timer
from ..tools.printer import Printer


class ParticleSolver(Solver):
    """
    Particular solver description.

    Link with differents numericals methods used.
    """
    def __init__(self, problem, t_end, dt,
                 ODESolver=RK3,
                 InterpolationMethod=None,
                 RemeshingMethod=None,
                 splittingMethod='strang',
                 timer=None,
                 io=None):
        """
        Create a solver description.

        @param problem : problen that solve this solver.
        @param t_end : Simulation final time.
        @param dt : Simulation time step.
        @param ODESolver : ODE solver method.
        @param InterpolationMethod : Interpolation method.
        @param RemeshingMethod : Remeshing method.
        @param splittingMethod : Splitting type.
        @param timer : Simulation time step manager.
        @param io : Simulation io manager.
        """
        Solver.__init__(self, problem)
        ## Problem to solve
        self.problem = problem
        ## Advection operator of the problem. Advection need special treatment in particle methods.
        self.advection = None
        self.stretch = None
        self.penal = None
        self.advec_scales = None
        self.diffusion = None
        self.poisson = None
        ## ODE Solver method.
        self.ODESolver = ODESolver
        ## Interpolation method.
        self.InterpolationMethod = InterpolationMethod
        ## Remeshing method.
        self.RemeshingMethod = RemeshingMethod
        ## Splitting Method
        self.splittingMethod = splittingMethod
        ## Is solver initialized
        self.isInitialized = False
        if timer is None:
            self.problem.setTimer(Timer(t_end, dt))
        else:
            self.problem.setTimer(timer)
        if io is None:
            self.problem.setIO(Printer())
        else:
            self.problem.setIO(io)
        for op in self.problem.operators:
            if isinstance(op, Transport):
                self.advection = op
            if isinstance(op, Stretching):
                self.stretch = op
            if isinstance(op, Penalization):
                self.penal = op
            if isinstance(op, Advec_scales):
                self.advec_scales = op
            if isinstance(op, Diffusion):
                self.diffusion = op
            if isinstance(op, Poisson):
                self.poisson = op
            #if isinstance(op, Velocity)
                #self.velocity = op
        #if self.advection is None:
        #    raise ValueError("Particular Solver : Cannot create from a problem with no Transport operator")

    def initialize(self):
        """
        Solver initialisation.
        Initialize a particle method solver regarding the problem.
        """
        self.p_position = ContinuousField(domain=self.problem.domains[0], name='ParticlePosition')
        self.p_scalar = ContinuousField(domain=self.problem.domains[0], name='ParticleScalar')
        self.problem.addVariable([self.p_position, self.p_scalar])
        ## Discretise domains
        for d in self.problem.domains:
            d.discretize(self.problem.topology.resolution)
        ## Discretise fields and initialize
        for v in self.problem.variables:
            v.discretize(self.problem.topology)
        print "=== Fields initialization ==="
        for v in self.problem.variables:
            v.initialize()
        print "==\n"
        ## Discretise operators
        for op in self.problem.operators:
            if op is self.advection:
                op.discretize(result_position=self.p_position, result_scalar=self.p_scalar, method=self.ODESolver)
            if op is self.stretch:
                op.discretize(topology=self.problem.topology, method=self.ODESolver)
            if op is self.penal:
                op.obstacle.discretize(self.problem.topology)
                op.discretize()
            if op is self.advec_scales:
                op.discretize(topology=self.problem.topology)
            if op is self.diffusion:
                op.discretize(topology=self.problem.topology)
            if op is self.poisson:
                op.discretize(topology=self.problem.topology)
            ## Velocity is only program on gpu for instance
            #elif op is self.velocity:
            #    op.discretize(result_position=self.p_position, result_scalar=self.p_scalar, method=self.ODESolver)
            else:
                op.discretize()
        if self.advection is not None:
            ## Create remeshing operator
            self.remeshing = Remeshing(partPositions=self.advection.discreteOperator.res_position,
                                    partScalar=self.advection.discreteOperator.res_scalar,
                                    resscal=self.advection.discreteOperator.scalar,
                                    method=self.RemeshingMethod)
        for op in self.problem.operators:
            if op.needSplitting:
                if op is self.advection:
                    index = self.problem.operators.index(op)
                    self.problem.operators[index] = Splitting([op, self.remeshing], dim=self.problem.topology.dim)
                else:
                    self.problem.operators[index] = Splitting([op], dim=self.problem.topology.dim)
        self.isInitialized = True

    def __str__(self):
        """ToString method"""
        s = " Particular solver "
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : ParticularSolver"
    print ParticularSolver.__doc__
