# -*- coding: utf-8 -*-
"""
@package Utils
RK4 method interface.
"""
from .integrator import ODESolver
import numpy as np


class RK4(ODESolver):
    """
    ODESolver implementation for solving an equation system with RK4 method.
    y'(t) = f(t,y)

    y(t_n+1)= y(t_n) + dt*y'[y(t_n)+dt/2*y'(y(t_n))].

    """
    def __init__(self, f=None, dim=3):
        """
        Constructor.

        @param f function f(t,y) : Right hand side of the equation to solve.
        @param dim : dimensions
        @param conditions : function to apply boundary conditions.
        """
        ODESolver.__init__(self, f)


    def integrate(self, f, t , dt , y):
        """
        Integration step for RK4 Method.
         up = f[t, Y(t,x,y,z)].

        @param y : function of position at time t.
        @param f : function of function of position at time t.
        @param t : current time.
        @param dt : time step.
        """
#K1 = dt *f[t,y(t,x,y,z)]
        K1 = dt * np.asarray(f(t,y))
        K2 = dt *np.asarray(f(t+dt/2.,y + K1[:,:,:,:]/2.))
        K3 = dt *np.asarray(f(t+dt/2.,y + K2[:,:,:,:]/2.))
        K4 = dt *np.asarray(f(t+dt,y + K3[:,:,:,:]))
#result = y(t,x,y,z)] + dt *f[t + dt/2 , y(t,x,y,z) + k1/2.]
        result = y + 1./6. * (K1[:,:,:,:] + 2.*K2[:,:,:,:] + 2.*K3[:,:,:,:]+ K4[:,:,:,:])
        return result

    def __str__(self):
        """ToString method"""
        return "RK4 Method"


if __name__ == "__main__":
    print __doc__
    print "- Provided class : RK4"
    print RK4.__doc__
