# -*- coding: utf-8 -*-
"""
@package physics
Compute forces
"""
from ..constants import *
import numpy as np
from ..operator.fct2op import Fct2Op
from ..operator.differentialOperator import DifferentialOperator
from ..operator.differentialOperator_d import DifferentialOperator_d
import mpi4py.MPI as MPI
import time


class Compute_forces(object):
    """
    Compute the forces according the Noca s formula
    """

    def __init__(self, topology, obstacle, boxMin= None, boxMax=None):
        """
        Constructor.
        @param topology : Local topology
        @param obstacle : obstacle 
        @param boxMin : Global minimum coordinates of the control volume 
        @param boxMax : Global maximum coordinates of the control volume  
        """
        self.topo = topology
        self.obstacle = obstacle
        self.boxMin = boxMin
        self.boxMax = boxMax
        self.dim = topology.dim
        self.res = topology.mesh.resolution
        self.step = topology.mesh.size
        self.coordMin = topology.mesh.origin + (topology.ghosts* self.step)
        self.coordMax = topology.mesh.end * self.step + topology.domain.origin - 2.* (topology.ghosts* self.step)
        self.ghosts = topology.ghosts
        self.periods = topology.periods
#        print 'res, dx, cMin, cMax, bmin, bMax', self.res, self.step, self.coordMin, self.coordMax, self.boxMin, self.boxMax
        self.compute_control_box()

    def compute_control_box(self) :
        """
        Compute indicator functions for the control box (including the obstacle)
        """
        # normal vectors definitions
        self.normal = np.zeros([self.dim*2, self.dim], dtype=PARMES_REAL)
        self.Up = 0
        self.Down = 1
        self.East = 2
        self.West = 3
        self.North = 4
        self.South = 5
        self.normal[self.Up][0] = 1.0
        self.normal[self.Down][0] = - 1.0
        self.normal[self.East][1] = 1.0
        self.normal[self.West][1] = - 1.0
        self.normal[self.North][2] = 1.0
        self.normal[self.South][2] = - 1.0

        # control box definition
        tmp_ind = np.zeros([self.dim, 2], dtype=PARMES_INTEGER)

        distMin = self.boxMin - self.coordMin
        distMax = self.boxMax - self.coordMax
        for i in xrange(self.dim):
            if (distMin[i]>=0. and distMax[i]<=0.): 
            # the control volume is included inside the local domain
                tmp_ind[i][0] = distMin[i] // self.step[i] # Down, West, South
                tmp_ind[i][1] = self.res[i] -2. * self.ghosts[i] + distMax[i] // self.step[i] -1# Up, East, North
            if (distMin[i]>=0. and self.boxMin[i]<= self.coordMax[i] and distMax[i]>=0.): 
            # only min corner of control volume is included inside the local domain
                tmp_ind[i][0] = distMin[i] // self.step[i]
                tmp_ind[i][1] = self.res[i] -2. * self.ghosts[i] - 1
            if (distMin[i]<=0. and self.boxMax[i]>= self.coordMin[i] and distMax[i]<=0.) : 
            # only max corner of control volume is included inside the local domain
                tmp_ind[i][1] = self.res[i] -2.*self.ghosts[i] + distMax[i] // self.step[i] -1
            if (distMin[i]<=0. and distMax[i]>=0.): 
            # the local domain is included inside the control volume
                tmp_ind[i][1] = self.res[i] - 2. * self.ghosts[i] - 1

            # otherwise, there is no volume control inside the local domain

        ind = np.zeros(self.dim*2, dtype=PARMES_INTEGER)
        ind[self.Down] = tmp_ind[0][0]
        ind[self.Up] = tmp_ind[0][1]
        ind[self.West] = tmp_ind[1][0]
        ind[self.East] = tmp_ind[1][1]
        ind[self.South] = tmp_ind[2][0]
        ind[self.North] = tmp_ind[2][1]
#        print 'ind_Down, ind_South', ind[self.Down], ind[self.Up]

        # Remove last point to integrate properly ... ?????????
#        ind[self.Up] = ind[self.Up] - 1
#        ind[self.East] = ind[self.East] - 1
#        ind[self.North] = ind[self.North] - 1
    
        # Count the number of points on each face and inside the domain
        nbPoints = np.zeros(2 * self.dim, dtype=PARMES_INTEGER)
        nbPoints[self.Down] = (ind[self.East] - ind[self.West] + 1) * (ind[self.North] - ind[self.South] + 1)
        nbPoints[self.Up] = (ind[self.East] - ind[self.West] + 1) * (ind[self.North] - ind[self.South] + 1)
        nbPoints[self.West] = (ind[self.Up] - ind[self.Down] + 1) * (ind[self.North] - ind[self.South] + 1)
        nbPoints[self.East] = (ind[self.Up] - ind[self.Down] + 1) * (ind[self.North] - ind[self.South] + 1)
        nbPoints[self.South] = (ind[self.East] - ind[self.West] + 1) * (ind[self.Up] - ind[self.Down] + 1)
        nbPoints[self.North] = (ind[self.East] - ind[self.West] + 1) * (ind[self.Up] - ind[self.Down] + 1)

#        print 'nb points plan Down', nbPoints[self.Down]

        # if there is no volume control inside the local domain => nbPoints=0 => forces are not computed
        if (distMin[0]<0. and distMax[0]<0. and self.boxMax[0]< self.coordMin[0]):
            nbPoints[self.Down] = 0
        if (distMin[0]>0. and distMax[0]>0. and self.boxMin[0]> self.coordMax[0]):
            nbPoints[self.Up] = 0
        if (distMin[1]<0. and distMax[1]<0. and self.boxMax[1]< self.coordMin[1]):
            nbPoints[self.West] = 0
        if (distMin[1]>0. and distMax[1]>0. and self.boxMin[1]> self.coordMax[1]):
            nbPoints[self.East] = 0
        if (distMin[2]<0. and distMax[2]<0. and self.boxMax[2]< self.coordMin[2]):
            nbPoints[self.South] = 0
        if (distMin[2]>0. and distMax[2]>0. and self.boxMin[2]> self.coordMax[2]):
            nbPoints[self.North] = 0

        boxDim = np.zeros(self.dim, dtype=PARMES_INTEGER)
        boxDim[0] = ind[self.Up] - ind[self.Down] + 1
        boxDim[1] = ind[self.East] - ind[self.West] + 1
        boxDim[2] = ind[self.North] - ind[self.South] + 1
        nbPointsBox = boxDim[0] * boxDim[1] * boxDim[2]

#        print 'nb points box', nbPointsBox

        tmp_box = np.zeros([nbPointsBox,self.dim], dtype=PARMES_INTEGER)
        coords = np.zeros(self.dim, dtype=PARMES_REAL)
        count_box = 0
        if(boxDim.all() >0):
            for k in xrange (ind[self.South],ind[self.North]+1): ## enlever les boucles !!
                coords[2] = self.coordMin[2] + k * self.step[2]
                for j in xrange (ind[self.West],ind[self.East]+1):
                    coords[1] = self.coordMin[1] + j * self.step[1]
                    for i in xrange (ind[self.Down],ind[self.Up]+1):
                        coords[0] = self.coordMin[0] + i * self.step[0]
                        dist = np.vdot(coords - self.obstacle.center,coords - self.obstacle.center) - self.obstacle.radius ** 2
                        if(dist >=0.0): # We are on or outside the sphere ...
                            tmp_box[count_box] = [i, j, k]
                            count_box = count_box + 1

#        print 'tmp_box', tmp_box, tmp_box.shape, count_box

        # local indices of points in the control box which are outside the obstacle
        self.chi_box = np.zeros([count_box,self.dim], dtype=PARMES_INTEGER)
        self.chi_box = tmp_box[0:count_box,:] ### PBM !!!

        self.chi_up = np.zeros([nbPoints[self.Up],self.dim], dtype=PARMES_INTEGER)
        self.chi_down = np.zeros([nbPoints[self.Down],self.dim], dtype=PARMES_INTEGER)
        self.chi_east = np.zeros([nbPoints[self.East],self.dim], dtype=PARMES_INTEGER)
        self.chi_west = np.zeros([nbPoints[self.West],self.dim], dtype=PARMES_INTEGER)
        self.chi_north = np.zeros([nbPoints[self.North],self.dim], dtype=PARMES_INTEGER)
        self.chi_south = np.zeros([nbPoints[self.South],self.dim], dtype=PARMES_INTEGER)

        # local indices of points located on the surfaces of the control box
        if(nbPoints[self.South]!=0.): # South boundary is in the domain
            count = 0
            for j in xrange (ind[self.West],ind[self.East] + 1):
                for i in xrange (ind[self.Down],ind[self.Up] + 1):
                    self.chi_south[count] = [i, j, ind[self.South]]
                    count = count + 1

        if(nbPoints[self.East]!=0.): # East boundary is in the domain
            count = 0
            for k in xrange (ind[self.South],ind[self.North] + 1):
                for i in xrange (ind[self.Down],ind[self.Up] + 1):
                    self.chi_east[count] = [i, ind[self.East], k]
                    count = count + 1

        if(nbPoints[self.Down]!=0.): # Down boundary is in the domain
            count = 0
            for k in xrange (ind[self.South],ind[self.North] + 1):
                for j in xrange (ind[self.West],ind[self.East] + 1):
                    self.chi_down[count] = [ind[self.Down], j, k]
                    count = count + 1

        if(nbPoints[self.North]!=0.): # North boundary is in the domain
            count = 0
            for j in xrange (ind[self.West],ind[self.East] + 1):
                for i in xrange (ind[self.Down],ind[self.Up] + 1):
                    self.chi_north[count] = [i, j, ind[self.North]]
                    count = count + 1

        if(nbPoints[self.West]!=0.): # West boundary is in the domain
            count = 0
            for k in xrange (ind[self.South],ind[self.North] + 1):
                for i in xrange (ind[self.Down],ind[self.Up] + 1):
                    self.chi_west[count] = [i, ind[self.West], k]
                    count = count + 1

        if(nbPoints[self.Up]!=0.): # Up boundary is in the domain
            count = 0
            for k in xrange (ind[self.South],ind[self.North] + 1):
                for j in xrange (ind[self.West],ind[self.East] + 1):
                    self.chi_up[count] = [ind[self.Up], j, k]
                    count = count + 1

        self.bufferForce = 0.
        # Compute coef used to determine drag coefficient : cD = coef.Fx = 2.Fx/rho.uinf**2.D
        uinf = 1.
        self.coef = 2. / (uinf ** 2 * PI * self.obstacle.radius ** 2)

    def apply(self, t, dt, velo, vort, Re):
        """
        Computation of the drag according to the "impulsion" formula presented by
        Noca et. al in Journal of Fluids and Structures, 1999
        """
        localForce = np.zeros(self.dim, dtype=PARMES_REAL)
        force = np.zeros(self.dim, dtype=PARMES_REAL)
        # computation of hessian matrices (d2_u1, d2_u2 and d2_u3) and jacobian matrix (d_u)
        hessian1, hessian2, hessian3 = Fct2Op(vort, velo, choice = 'hessianU', topology=self.topo).apply(None,None)
        self.jacobian, maxgersh = DifferentialOperator_d(vort, velo, choice='gradV', topology=self.topo).apply()
        # computation of the components of Div(T) where T = 1/Re * (Nabla(U) + Nabla(U)T)
        self.a = 1. / Re * (2. * hessian1[0] + hessian1[4] + hessian2[1] + hessian1[8] + hessian3[2])
        self.b = 1. / Re * (hessian2[0] + hessian1[1] + 2. * hessian2[4] + hessian2[8] + hessian3[5])
        self.c = 1. / Re * (hessian3[0] + hessian1[6] + hessian3[4] + hessian2[7] + 2. * hessian3[8])

        # Downstream and upstream surfaces
        dsurf = self.step[1] * self.step[2]
        localForce = self.integrateOnSurface(localForce, velo, vort, self.chi_down, self.normal[self.Down][:], Re, dsurf)
        localForce = self.integrateOnSurface(localForce, velo, vort, self.chi_up, self.normal[self.Up][:], Re, dsurf)
        # East and West surfaces
        dsurf = self.step[0] * self.step[2]
        localForce = self.integrateOnSurface(localForce, velo, vort, self.chi_east, self.normal[self.East][:], Re, dsurf)
        localForce = self.integrateOnSurface(localForce, velo, vort, self.chi_west, self.normal[self.West][:], Re, dsurf)
        # North and south surfaces
        dsurf = self.step[0] * self.step[1]
        localForce = self.integrateOnSurface(localForce, velo, vort, self.chi_south, self.normal[self.South][:], Re, dsurf)
        localForce = self.integrateOnSurface(localForce, velo, vort, self.chi_north, self.normal[self.North][:], Re, dsurf)
        # over the volume
        dvol = self.step[0] * self.step[1] * self.step[2]
        localForce = self.integrateOnBox(localForce, vort, self.chi_box, dvol, dt)

        localForce = localForce * self.coef
        comm = MPI.COMM_WORLD
        comm.Reduce([localForce, MPI.DOUBLE], [force, MPI.DOUBLE], op=MPI.SUM, root=0)

        return force 

    def integrateOnBox(self, force, vort, chi, dvol, dt):
        """
        Return -1/(dim-1)*d/dt int_over_control_box coord X vorticity 
        """
        fact = - dvol / ((self.dim - 1) * dt) 
        int1 = np.zeros(self.dim, dtype=PARMES_REAL)
        # For all points in the box
        for ind in xrange (chi.shape[1]):
            # adjust indices for vector fields with ghost points
            i = chi[ind][0] + self.ghosts[0]
            j = chi[ind][1] + self.ghosts[1]
            k = chi[ind][2] + self.ghosts[2]
            # coordinates of the current point
            coords = self.coordMin + chi[ind] * self.step
            # part1 of the force
            int1 = int1 + np.vdot(coords,vort[i,j,k])

        force = force + fact * (int1 - self.bufferForce) # 1st order time integration
        self.bufferForce = int1 # Save for next time step ...

        return force 

    def integrateOnSurface(self, force, velo, vort, chi, NormalVec, Re, dsurf):
        """
        Compute integrals on surface to calculate forces acting on the body.
        See (2.1) of Noca 1999 or (52) of Plouhmans 2002 
        Integrals on the obstacle are neglected. 
        """

        fact = 1. / (self.dim - 1)
        int1 = np.zeros(self.dim, dtype=PARMES_REAL)
        int2 = np.zeros(self.dim, dtype=PARMES_REAL)
        X_n_DivT = np.zeros(self.dim, dtype=PARMES_REAL)
        n_T = np.zeros(self.dim, dtype=PARMES_REAL)
        # For each point of the current plane
        for ind in xrange (chi.shape[0]):
            # adjust indices for vector fields with ghost points
            i = chi[ind][0] + self.ghosts[0]
            j = chi[ind][1] + self.ghosts[1]
            k = chi[ind][2] + self.ghosts[2]
            # part1 = 1/2(velocity.velocity)n - (n.velocity)velocity - 1/(dim-1)((n.velocity)(coord X vorticity) + 1/(dim-1)(n.vorticity)(coord X velocity)

            # (velocity.velocity)
            u_u = np.vdot(velo[i,j,k],velo[i,j,k])
            # normal.velocity
            n_u = np.vdot(velo[i,j,k],NormalVec[:])
            # normal.vorticity
            n_w = np.vdot(vort[i,j,k],NormalVec[:])
            # coordinates of the current point
            coords = self.coordMin[:] + chi[ind] * self.step[:]
            # part1 of the force
            int1 = int1 + 0.5 * u_u * NormalVec[:] - n_u * velo[i,j,k] \
                - fact * n_u * np.cross(coords,vort[i,j,k]) \
                + fact * n_w * np.cross(coords,velo[i,j,k])

            # part 2 of the force : part 2 = 1/(dim-1)(X^(n^Div(T)) + n.T
            if (chi.all() == self.chi_down.all()): # normalVec = (-1,0,0)
                X_n_DivT[0] = - coords[1] * self.b[i,j,k] - coords[2] * self.c[i,j,k]
                X_n_DivT[1] = coords[0] * self.b[i,j,k]
                X_n_DivT[2] = coords[0] * self.c[i,j,k]
                n_T[0] = - 1. / Re * 2 * self.jacobian[0][i,j,k]
                n_T[1] = - 1. / Re * (self.jacobian[3][i,j,k] + self.jacobian[1][i,j,k])
                n_T[2] = - 1. / Re * (self.jacobian[6][i,j,k] + self.jacobian[2][i,j,k])

            if (chi.all() == self.chi_up.all()): # normalVec = (1,0,0)
                X_n_DivT[0] = coords[1] * self.b[i,j,k] + coords[2] * self.c[i,j,k]
                X_n_DivT[1] = - coords[0] * self.b[i,j,k]
                X_n_DivT[2] = - coords[0] * self.c[i,j,k]
                n_T[0] = 1. / Re * 2 * self.jacobian[0][i,j,k]
                n_T[1] = 1. / Re * (self.jacobian[3][i,j,k] + self.jacobian[1][i,j,k])
                n_T[2] = 1. / Re * (self.jacobian[6][i,j,k] + self.jacobian[2][i,j,k])

            if (chi.all() == self.chi_west.all()): # normalVec = (0,-1,0)
                X_n_DivT[0] = coords[1] * self.a[i,j,k]
                X_n_DivT[1] = - coords[0] * self.a[i,j,k] - coords[2] * self.c[i,j,k]
                X_n_DivT[2] = coords[1] * self.c[i,j,k]
                n_T[0] = - 1. / Re * (self.jacobian[1][i,j,k] + self.jacobian[3][i,j,k])
                n_T[1] = - 1. / Re * 2 * self.jacobian[4][i,j,k]
                n_T[2] = - 1. / Re * (self.jacobian[7][i,j,k] + self.jacobian[5][i,j,k])

            if (chi.all() == self.chi_east.all()): # normalVect = (0,1,0)
                X_n_DivT[0] = - coords[1] * self.a[i,j,k]
                X_n_DivT[1] = coords[0] * self.a[i,j,k] + coords[2] * self.c[i,j,k]
                X_n_DivT[2] = - coords[1] * self.c[i,j,k]
                n_T[0] = 1. / Re * (self.jacobian[1][i,j,k] + self.jacobian[3][i,j,k])
                n_T[1] = 1. / Re * 2 * self.jacobian[4][i,j,k]
                n_T[2] = 1. / Re * (self.jacobian[7][i,j,k] + self.jacobian[5][i,j,k])

            if (chi.all() == self.chi_south.all()): # normalVec = (0,0,-1)
                X_n_DivT[0] = coords[2] * self.a[i,j,k]
                X_n_DivT[1] = coords[2] * self.b[i,j,k]
                X_n_DivT[2] = - coords[0] * self.a[i,j,k] - coords[1] * self.b[i,j,k]
                n_T[0] = - 1. / Re * (self.jacobian[2][i,j,k] + self.jacobian[6][i,j,k])
                n_T[1] = - 1. / Re * (self.jacobian[5][i,j,k] + self.jacobian[7][i,j,k])
                n_T[2] = - 1. / Re * 2 * self.jacobian[8][i,j,k] 

            if (chi.all() == self.chi_north.all()): # normalVec = (0,0,1)
                X_n_DivT[0] = - coords[2] * self.a[i,j,k]
                X_n_DivT[1] = - coords[2] * self.b[i,j,k]
                X_n_DivT[2] = coords[0] * self.a[i,j,k] + coords[1] * self.b[i,j,k]
                n_T[0] = 1. / Re * (self.jacobian[2][i,j,k] + self.jacobian[6][i,j,k])
                n_T[1] = 1. / Re * (self.jacobian[5][i,j,k] + self.jacobian[7][i,j,k])
                n_T[2] = 1. / Re * 2 * self.jacobian[8][i,j,k] 

            int2 = int2 + fact * X_n_DivT + n_T
        # Product with element of surface and sum to the total (input) force
        result = force + (int1 + int2) * dsurf

        return result


    def __str__(self):
        s = "Compute_forces. "
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Compute_forces"
    print Compute_forces.__doc__
