"""
@package parmepy.problem.problem

Problem representation.
"""
from ..particular_solvers.basic import ParticleSolver
from ..particular_solvers.gpu import GPUParticleSolver


class Problem():
    """
    Problem representation.

    Contains several operators that apply on variables. Variables are defined on different domains.\n
    The problem contains a solver that define how the problem is solved.\n
    A problem is based on a topology that deploys its resolution over MPI processus.\n
    Finally, the problem handle outputs thanks to a io object.
    """

    def __init__(self, topology, operators, params=None):
        """
        Create a transport problem instance.

        @param topology : underlying topology. \todo make topology transparent to user.
        @param operators : list of operators.
        @param params : unused.
        """
        ## Problem domains
        self.domains = []
        ## Prolem variables
        self.variables = []
        ## Problem operators
        self.operators = operators
        ## Problem topology
        self.topology = topology
        ## Solver for problem
        self.solver = None
        ## Computes time step and manage iterations
        self.timer = None
        ## IO manager
        self.io = None
        ## Timings informations
        self.timings_info = ["", ""]
        for op in self.operators:
            self.addVariable(op.variables)
        for v in self.variables:
            self.addDomain(v.domain)

    def setTimer(self, t):
        """
        Set timer.
        @param t : timer.
        """
        self.timer = t

    def setIO(self, io):
        """
        Set IO.
        @param io : io interface.
        """
        self.io = io

    def setSolver(self, t_end, dt, solver_type='gpu', **kwargs):
        """
        Set solver.
        @param t_end : simulation final time.
        @param dt : simulation time step.
        @param solver_type : type of solver to use. Default : gpu solver.
        @param **kwargs : dynamic dictionary of arguments specific to solver.
        """
        if solver_type == 'basic':
            self.solver = ParticleSolver(self, t_end, dt, **kwargs)
        elif solver_type == 'gpu':
            self.solver = GPUParticleSolver(self, t_end, dt, **kwargs)
        else:
            raise ValueError("Unknown solver type : " + str(solver_type))

    def initSolver(self):
        """Initialize solver."""
        if self.solver is not None:
            self.solver.initialize()

    def solve(self):
        """
        Solve problem.

        Performs simulations iterations by calling each operators of the list until timer end.\n
        At end of time step, call an io step.\n
        Displays timings at simulation end.
        """
        print "\n\n Start solving ..."
        if not self.solver.isInitialized:
            self.initSolver()
        self.io.step()
        while not self.timer.end:
            if ( self.topology.rank == 0 ) :
                print "==== Iteration : {0:3d}   t={1:6.3f} ====".format(self.timer.ite + 1, self.timer.t + self.timer.dt)
            for op in self.operators:
                op.apply(self.timer.t, self.timer.dt)
            self.timer.step()
            self.io.step()
        print "\n\n End solving\n"
        print "=== Timings ==="
        if ( self.topology.rank == 0 ) :
            print "IO total", self.io.compute_time
            self.timings_info[0] += "\"IO\" "
            self.timings_info[1] += str(self.io.compute_time) + " "
            for op in self.operators:
                op.printComputeTime()
                self.timings_info[0] += op.timings_info[0]
                self.timings_info[1] += op.timings_info[1]
            print "\n"
            print "===\n"

    def addVariable(self, cVariable):
        """
        Add an continuous variables to the operator.
        Also add variables' domains to the operator.

        @param cVariable ContinuousVariable.ContinuousVariable : variables to add.
        """
        try:
            for v in cVariable:
                if self.variables.count(v) == 0:
                    self.variables.append(v)
        except (TypeError):
            if self.variables.count(cVariable) == 0:
                self.variables.append(cVariable)

    def addDomain(self, cDomain):
        """
        Add an domain to the operator.

        @param cDomain : domain to add
        """
        try:
            for d in cDomain:
                if self.domains.count(d) == 0:
                    self.domains.append(d)
        except (TypeError):
            if self.domains.count(cDomain) == 0:
                self.domains.append(cDomain)

    def addOperator(self, op, index=None):
        """
        Add operator to problem.

        @param op : operator.
        @param index : operator index.
        """
        if not op in self.operators:
            if index is None:
                self.operators.append(op)
            else:
                self.operators.insert(index, op)

    def __str__(self):
        """ToString method"""
        s = "Problem based on\n"
        s += str(self.topology)
        s += "with following operators : \n"
        for op in self.operators:
            s += str(op)
        s += "working on fields : \n"
        for f in self.variables:
            s += str(f)
        s += "defined on domains : \n"
        for d in self.domains:
            s += str(d)
        s += "\nsolved by : \n" + str(self.solver)

        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Problem"
    print ContinuousTransportProblem.__doc__
