"""
Module for testing parmepy.field.discrete
"""

import unittest
import parmepy as pp
import numpy as np


class ScalarFieldTestCase(unittest.TestCase):
    def setUp(self):
        self.b = pp.Box()
        self.nbElem = [10, 10, 10]
        topo = pp.CartesianTopology(domain=self.b,
                                    resolution=self.nbElem,
                                    dim=len(self.nbElem))
        self.formula = lambda x, y, z: x * y * z
        field_scalar = pp.fields.analytical.AnalyticalField(self.b,
                                                            formula=self.formula,
                                                            name="test_field_scalar")
        self.ds, self.ds_id = field_scalar.discretize(topo)

    def tearDown(self):
        pass

    def test_constructor(self):
        """Test ScalarField construction"""
        np.testing.assert_array_equal(self.ds.data, np.zeros(tuple(self.nbElem), dtype=pp.constants.PARMES_REAL))
        self.assertEqual(self.ds.data.shape, tuple(self.nbElem))
        self.assertTrue(not self.ds.contains_data)

    def test_initialization(self):
        """Scalar initialization"""
        self.ds.initialize(self.formula)
        true_res = np.zeros_like(self.ds.data)
        for i in np.arange(10):
            for j in np.arange(10):
                for k in np.arange(10):
                    true_res[i, j, k] = self.formula(i / 10., j / 10., k / 10.)
        np.testing.assert_array_almost_equal(self.ds.data, true_res)


class VectorFieldTestCase(unittest.TestCase):
    def setUp(self):
        self.b = pp.Box()
        self.nbElem = [10, 10, 10]
        topo = pp.CartesianTopology(domain=self.b,
                                    resolution=self.nbElem,
                                    dim=len(self.nbElem))
        self.formula = lambda x, y, z: (y - x, - x * z - y, x * y - z)
        field_vector = pp.fields.analytical.AnalyticalField(self.b,
                                                            formula=self.formula,
                                                            name="test_field_vector", vector=True)
        self.dv, self.dv_id = field_vector.discretize(topo)

    def tearDown(self):
        pass

    def test_constructor(self):
        """Test VectorField construction"""
        self.assertTrue(isinstance(self.dv.data, list))
        self.assertEqual(len(self.dv.data), 3)
        for df in self.dv.data:
            np.testing.assert_array_equal(df, np.zeros(tuple(self.nbElem), dtype=pp.constants.PARMES_REAL))
            self.assertEqual(df.shape, tuple(self.nbElem))
        self.assertTrue(not self.dv.contains_data)

    def test_initialization(self):
        """Vector initialization"""
        self.dv.initialize(self.formula)
        true_res_x = np.zeros_like(self.dv.data[0])
        true_res_y = np.zeros_like(self.dv.data[1])
        true_res_z = np.zeros_like(self.dv.data[2])
        for i in np.arange(10):
            for j in np.arange(10):
                for k in np.arange(10):
                    true_res_x[i, j, k], true_res_y[i, j, k], true_res_z[i, j, k] = self.formula(i / 10., j / 10., k / 10.)
        np.testing.assert_array_almost_equal(self.dv.data[0], true_res_x)
        np.testing.assert_array_almost_equal(self.dv.data[1], true_res_y)
        np.testing.assert_array_almost_equal(self.dv.data[2], true_res_z)

    def test_elementAcces(self):
        """Vector element access"""
        self.dv.initialize(self.formula)
        true_res_x = np.zeros_like(self.dv.data[0])
        true_res_y = np.zeros_like(self.dv.data[1])
        true_res_z = np.zeros_like(self.dv.data[2])
        self.assertEqual(self.dv[:, :, :].shape, (3, 10, 10, 10))
        self.assertEqual(self.dv[0][:, :, :].shape, (10, 10, 10))
        self.assertEqual(self.dv[0,0,0].shape, (3,))
        for i in np.arange(10):
            for j in np.arange(10):
                for k in np.arange(10):
                    true_res_x[i, j, k], true_res_y[i, j, k], true_res_z[i, j, k] = self.formula(i / 10., j / 10., k / 10.)
                    self.assertAlmostEqual(self.dv[0][i, j, k], true_res_x[i, j, k])
                    self.assertAlmostEqual(self.dv[1, i, j, k], true_res_y[i, j, k])
                    np.testing.assert_array_almost_equal(self.dv[i, j, k], [true_res_x[i, j, k], true_res_y[i, j, k], true_res_z[i, j, k]])
        np.testing.assert_array_almost_equal(self.dv[0], true_res_x)
        np.testing.assert_array_almost_equal(self.dv[1], true_res_y)
        np.testing.assert_array_almost_equal(self.dv[2], true_res_z)
        np.testing.assert_array_almost_equal(self.dv[0][:,1,2], true_res_x[:,1,2])
        np.testing.assert_array_almost_equal(self.dv[0,:,1,2], true_res_x[:,1,2])
        self.dv[1,2,3] = [1., 2., 3.]
        np.testing.assert_array_almost_equal( self.dv[1,2,3], [1., 2., 3.])
        self.dv[0,1,:] = 5.
        true_res = [np.ones_like(self.dv.data[0][0,1,:])*5.,
                    np.ones_like(self.dv.data[0][0,1,:])*5.,
                    np.ones_like(self.dv.data[0][0,1,:])*5.]
        np.testing.assert_array_almost_equal( self.dv[0,1,:], true_res)
        self.dv[0][...,9] = -1.
        np.testing.assert_array_almost_equal(self.dv[0][...,9], np.ones_like(true_res_x[...,9])*-1.)


if __name__ == "__main__":
    unittest.TextTestRunner(verbosity=2).run(
        unittest.TestLoader().loadTestsFromTestCase(ScalarFieldTestCase)
        )
    unittest.TextTestRunner(verbosity=2).run(
        unittest.TestLoader().loadTestsFromTestCase(VectorFieldTestCase)
        )
