"""
Module for testing parmepy.operator.transport_d
"""

import unittest
import parmepy as pp
import numpy as np
import pyopencl as cl


class GPUParticleSolverTestCase(unittest.TestCase):

    def setUp_OpenCL_basic(self):
        self.b = pp.Box()
        self.nbElem = [10, 10, 10]
        self.topo = pp.CartesianTopology(domain=self.b,
                                         resolution=self.nbElem,
                                         dim=len(self.nbElem))
        build_options = "-cl-single-precision-constant -cl-opt-disable"
        build_options += " -D WIDTH=10"
        build_options += " -D WGN=10"
        build_options += " -D PADDING=0"
        build_options += " -D BASIC=1"
        self.prg_basic = cl.Program(self.ctx, self.gpu_src).build(build_options)
        self.p_positions = pp.fields.continuous.ContinuousField(self.b,
                                                                name="test_field_positions")
        self.true_p_positions = pp.fields.analytical.AnalyticalField(self.b,
                                                                     formula=lambda x,y,z:x,
                                                                         name="test_field_true_positions")
        self.p_vitesse = pp.fields.analytical.AnalyticalField(self.b,
                                                              formula=lambda x, y, z: 0.,
                                                              name="test_field_velocity")
        self.pos, self.pos_id = self.p_positions.discretize(self.topo)
        self.velo, self.velo_id = self.p_vitesse.discretize(self.topo)
        self.true_pos, self.velo_id = self.true_p_positions.discretize(self.topo)
        self.true_p_positions.initialize()
        self.pos.data = np.asarray(self.pos.data, dtype=pp.constants.PARMES_REAL_GPU, order='F')
        self.velo.data = np.asarray(self.velo.data, dtype=pp.constants.PARMES_REAL_GPU, order='F')
        self.pos.gpu_data = cl.Buffer(self.ctx, cl.mem_flags.READ_WRITE,
                                      size=self.pos.data.nbytes)
        self.velo.gpu_data = cl.Buffer(self.ctx, cl.mem_flags.READ_WRITE,
                                       size=self.velo.data.nbytes)
        cl.enqueue_copy(self.queue, self.velo.gpu_data, self.velo.data)


    def setUp_OpenCL_opt(self):
        self.b = pp.Box()
        self.nbElem = [256, 256, 256]
        self.topo = pp.CartesianTopology(domain=self.b,
                                         resolution=self.nbElem,
                                         dim=len(self.nbElem))
        build_options = "-cl-single-precision-constant -cl-opt-disable"
        build_options += " -D WIDTH=256"
        build_options += " -D WGN=64"
        build_options += " -D PADDING=0"
        self.prg_opt = cl.Program(self.ctx, self.gpu_src).build(build_options)
        self.p_positions = pp.fields.continuous.ContinuousField(self.b,
                                                                name="test_field_positions")
        self.true_p_positions = pp.fields.analytical.AnalyticalField(self.b,
                                                                     formula=lambda x,y,z:x,
                                                                         name="test_field_true_positions")
        self.p_vitesse = pp.fields.analytical.AnalyticalField(self.b,
                                                              formula=lambda x, y, z: 0.,
                                                              name="test_field_velocity")
        self.pos, self.pos_id = self.p_positions.discretize(self.topo)
        self.velo, self.velo_id = self.p_vitesse.discretize(self.topo)
        self.true_pos, self.velo_id = self.true_p_positions.discretize(self.topo)
        self.true_p_positions.initialize()
        self.pos.data = np.asarray(self.pos.data, dtype=pp.constants.PARMES_REAL_GPU, order='F')
        self.velo.data = np.asarray(self.velo.data, dtype=pp.constants.PARMES_REAL_GPU, order='F')
        self.pos.gpu_data = cl.Buffer(self.ctx, cl.mem_flags.READ_WRITE,
                                      size=self.pos.data.nbytes)
        self.velo.gpu_data = cl.Buffer(self.ctx, cl.mem_flags.READ_WRITE,
                                       size=self.velo.data.nbytes)
        cl.enqueue_copy(self.queue, self.velo.gpu_data, self.velo.data)

    def setUp(self):
        try:
            self.platform = cl.get_platforms()[0]
            self.device = self.platform.get_devices(cl.device_type.GPU)[0]
            self.ctx = cl.Context([self.device])
            self.queue = cl.CommandQueue(self.ctx, properties=cl.command_queue_properties.PROFILING_ENABLE)
            f = open(pp.constants.GPU_SRC, 'r')
            self.gpu_src = "".join(f.readlines())
            f.close()
            self.test = True
            print "Testing on " + self.device.name
        except Exception:
            self.test = False
            print "No tests because no GPU"

    def tearDown(self):
        pass

    def test_advection_basic_kernel(self):
        """Test advection basic kernel"""
        if self.test:
            self.setUp_OpenCL_basic()
            self.queue.finish()
            self.prg_basic.advection(self.queue, (10, 10, 10), None,
                               self.velo.gpu_data, self.pos.gpu_data,
                               pp.constants.PARMES_REAL_GPU(0.1),
                               pp.constants.PARMES_REAL_GPU(0.),
                               pp.constants.PARMES_REAL_GPU(1/10.))
            self.pos.data[...] = 0.
            self.queue.finish()
            cl.enqueue_copy(self.queue, self.pos.data, self.pos.gpu_data)
            self.queue.finish()
            np.testing.assert_array_almost_equal(self.pos.data, self.true_pos.data)
            self.velo.data[...] = 1.
            cl.enqueue_copy(self.queue, self.velo.gpu_data, self.velo.data)
            self.queue.finish()
            self.prg_basic.advection(self.queue, (10, 10, 10), None,
                               self.velo.gpu_data, self.pos.gpu_data,
                               pp.constants.PARMES_REAL_GPU(0.1),
                               pp.constants.PARMES_REAL_GPU(0.),
                               pp.constants.PARMES_REAL_GPU(1/10.))
            self.pos.data[...] = 0.
            self.queue.finish()
            cl.enqueue_copy(self.queue, self.pos.data, self.pos.gpu_data)
            self.queue.finish()
            np.testing.assert_array_almost_equal(self.pos.data, self.true_pos.data+.1)

    def test_advection_opt_kernel(self):
        """Test advection basic kernel"""
        if self.test:
            self.setUp_OpenCL_opt()
            self.queue.finish()
            self.prg_opt.advection(self.queue, (64, 256, 256), (64, 1, 1),
                               self.velo.gpu_data, self.pos.gpu_data,
                               pp.constants.PARMES_REAL_GPU(0.1),
                               pp.constants.PARMES_REAL_GPU(0.),
                               pp.constants.PARMES_REAL_GPU(1/256.))
            self.pos.data[...] = 0.
            self.queue.finish()
            cl.enqueue_copy(self.queue, self.pos.data, self.pos.gpu_data)
            self.queue.finish()
            np.testing.assert_array_almost_equal(self.pos.data, self.true_pos.data)
            self.velo.data[...] = 1.
            cl.enqueue_copy(self.queue, self.velo.gpu_data, self.velo.data)
            self.queue.finish()
            self.prg_opt.advection(self.queue, (64, 256, 256), (64, 1, 1),
                               self.velo.gpu_data, self.pos.gpu_data,
                               pp.constants.PARMES_REAL_GPU(0.1),
                               pp.constants.PARMES_REAL_GPU(0.),
                               pp.constants.PARMES_REAL_GPU(1/256.))
            self.pos.data[...] = 0.
            self.queue.finish()
            cl.enqueue_copy(self.queue, self.pos.data, self.pos.gpu_data)
            self.queue.finish()
            np.testing.assert_array_almost_equal(self.pos.data, self.true_pos.data+.1)


if __name__ == "__main__":
    unittest.TextTestRunner(verbosity=2).run(
        unittest.TestLoader().loadTestsFromTestCase(GPUParticleSolverTestCase)
        )
