"""
@package parmepy.tools.gpu_data_transfer

GPU Memory transfering methods.
"""
from ..constants import *
import pyopencl as cl


def hostToDevice(queue, discreteField):
    """
    Host to device method.

    @param queue : OpenCL queue.
    @param discreteField : Variable to transfer.

    Performs a direct OpenCL copy from numpy arrays to OpenCL Buffers.\n
    Arrange memory on device so that vector components are contiguous in the direction of the component.\n
    Example : A 3D vector field F(x,y,z) is made up of 3 OpenCL Buffers Fx, Fy, Fz. The memory layout is :
      - Fx : x-major ordering. On device, Fx[i + j*WIDTH + k*WIDTH*HEIGHT] access to Fx(i,j,k)
      - Fy : y-major ordering. On device, Fy[i + j*WIDTH + k*WIDTH*HEIGHT] access to Fy(j,i,k)
      - Fz : z-major ordering. On device, Fz[i + j*WIDTH + k*WIDTH*HEIGHT] access to Fz(k,i,j)
    """
    print "host->device :",
    data = 0
    time = 0.
    evt = [None, None, None]
    if discreteField.vector:
        evt[0] = cl.enqueue_copy(queue, discreteField.gpu_data[XDIR],
                              discreteField.data[XDIR].ravel(order=ORDER))
        data += discreteField.gpu_data[XDIR].size
        evt[1] = cl.enqueue_copy(queue, discreteField.gpu_data[YDIR],
                              discreteField.data[YDIR].swapaxes(0, 1).ravel(order=ORDER))
        data += discreteField.gpu_data[YDIR].size
        if len(discreteField.data) == 3:
            evt[2] = cl.enqueue_copy(queue, discreteField.gpu_data[ZDIR],
                                  discreteField.data[ZDIR].swapaxes(0, 1).swapaxes(0, 2).ravel(order=ORDER))
            data += discreteField.data[ZDIR].nbytes
    else:
        evt[0] = cl.enqueue_copy(queue, discreteField.gpu_data, discreteField.data.ravel(order=ORDER))
        data += discreteField.gpu_data.size
    queue.finish()
    for e in evt:
        if e is not None:
            time += (e.profile.end - e.profile.start) * 1e-9
    print data, "Bytes transfered at {0:.3f} GBytes/sec".format(data / (time * 1024 ** 3))
    return data, time


def deviceToHost(queue, discreteField):
    """
    Device to host method.

    @param queue : OpenCL queue.
    @param discreteField : Variable to transfer.

    Performs a direct OpenCL copy from OpenCL Buffers to numpy arrays.\n
    As memory layout is arranged on device, not only a copy is performed but also transpositions to have numpy arrays consistent to each other.
    """
    print "device->host :",
    data = 0
    time = 0.
    evt = [None, None, None]
    if discreteField.vector:
        # Use of discreteField.data[XDIR] as temporary array \todo Need cleaner solution
        # Get ZDIR component
        if len(discreteField.data) == 3:
            evt[2] = cl.enqueue_copy(queue, discreteField.data[XDIR],
                                  discreteField.gpu_data[ZDIR])
            discreteField.data[ZDIR] = discreteField.data[XDIR].swapaxes(0,2).swapaxes(0,1)
            data += discreteField.data[ZDIR].nbytes
        # Get YDIR component
        evt[1] = cl.enqueue_copy(queue, discreteField.data[XDIR],
                              discreteField.gpu_data[YDIR])
        discreteField.data[YDIR] = discreteField.data[XDIR].swapaxes(0,1)
        data += discreteField.gpu_data[YDIR].size
        # Get XDIR component
        evt[0] = cl.enqueue_copy(queue, discreteField.data[XDIR],
                              discreteField.gpu_data[XDIR])
        data += discreteField.gpu_data[XDIR].size
    else:
        evt[0] = cl.enqueue_copy(queue, discreteField.data,
                              discreteField.gpu_data)
        data += discreteField.gpu_data.size
    queue.finish()
    for e in evt:
        if e is not None:
            time += (e.profile.end - e.profile.start) * 1e-9
    print data, "Bytes transfered at {0:.3f} GBytes/sec".format(data / (time * 1024 ** 3))
    return data, time
