"""
@package parmepy.tools.printer

Classes for handling ouputs.
"""
from ..constants import *
import evtk.hl as evtk
import time


class Printer():
    """
    Basic printer.

    Performs outputs in VTK images.
    """

    def __init__(self, fields=[], frequency=0, outputPrefix='./out_'):
        """
        Create a results printer for given fields, filename prefix (relative path) and an output frequency.

        @param fields : Fields to output.
        @param frequency : output file producing frequency.
        @param outputPrefix : file name prefix, contains relative path.
        """
        ## Printer frequency
        self.freq = frequency
        ## Filename prefix
        self.outputPrefix = outputPrefix
        ## Fields to output
        self.fields = fields
        ## Iterations counter
        self.ite = 0
        ## Method to collect data in case of distributed data
        self.get_data_method = None
        if self.freq != 0:
            ## Printer core method
            self.step = self._printStep
        else:
            self.step = self._passStep
        ## Printing compute time
        self.compute_time = 0.

    def _build_vtk_dict(self):
        """Build a dictionary from fields to VTK image."""
        res = {}
        for f in self.fields:
            for df in f.discreteField:
                if f.vector:
                    for d in xrange(len(df.data)):
                        if len(df.data[d].shape) == 2:
                            res[df.name + S_DIR[d]] = np.expand_dims(df.data[d], axis=2)
                        else:
                            res[df.name + S_DIR[d]] = df.data[d]
                else:
                    if len(df.data.shape) == 2:
                        res[df.name] = np.expand_dims(df.data, axis=2)
                    else:
                        res[df.name] = df.data
        return res

    def set_get_data_method(self, method):
        """Set data collect method."""
        if callable(method):
            self.get_data_method = method
        else:
            raise ValueError("Cannot set non callable method to get data to print. Given method : " + str(method))

    def _passStep(self):
        """A code method that do nothing."""
        pass

    def _printStep(self):
        """
        Printer core method.

        Write fields data to file on necessary.\n
        VTK images are writed. If fails, classical ascii output is performed.
        """
        t = time.time()
        if (self.ite % self.freq) == 0:
            print "== IO"
            if self.get_data_method is not None:
                for f in self.fields:
                    for df in f.discreteField:
                        if not df.contains_data:
                            print "Data transfer : ", f.name, ":",
                            self.get_data_method(df)
            filename = self.outputPrefix + "results_{0:05d}.dat".format(self.ite)
            print "Print file : " + filename
            ## VTK output \todo: Need fix in 2D, getting an IOError.
            try:
                evtk.imageToVTK(filename, pointData=self._build_vtk_dict())
                print "not yet implemented"
            except IOError:
                ## Standard output
                f = open(filename, 'w')
                shape = self.fields[0].domain.discreteDomain.resolution
                step = self.fields[0].domain.discreteDomain.step
                if len(shape) == 2:
                    for i in xrange(shape[0]-1):
                        for j in xrange(shape[1]-1):
                            f.write("{0:8.12} {1:8.12} ".format(i * step[0], j * step[1]))
                            for field in self.fields:
                                if field.vector:
                                    f.write("{0:8.12} {1:8.12} ".format(field.discreteField[0][0][i, j],
                                                                        field.discreteField[0][1][i, j]))
                                else:
                                    f.write("{0:8.12} ".format(field.discreteField[0][i, j]))
                            f.write("\n")
                else:
                    for i in xrange(shape[0]-1):
                        for j in xrange(shape[1]-1):
                            for k in xrange(shape[2]-1):
                                f.write("{0:8.12} {1:8.12} {2:8.12} ".format(i * step[0], j * step[1], k * step[2]))
                                for field in self.fields:
                                    if field.vector:
                                        f.write("{0:8.12} {1:8.12} {2:8.12} ".format(field.discreteField[0][0][i, j, k],
                                                                                     field.discreteField[0][1][i, j, k],
                                                                                     field.discreteField[0][2][i, j, k]))
                                    else:
                                        f.write("{0:8.12} ".format(field.discreteField[0][i, j, k]))
                                f.write("\n")
                f.close()
                print "==\n"
        self.ite += 1
        self.compute_time += (time.time() - t)


if __name__ == "__main__":
    print __doc__
    print "- Provided class : " + providedClass
    print eval(providedClass).__doc__
