# -*- coding: utf-8 -*-
import unittest

from src.Problem.ContinuousTransportProblem import *
from src.Domain.Box import *
from src.Variable.AnalyticalVariable import *
from src.Operator.Advection import *
from src.Utils.ParticularSolver import *
from src.Utils.ForwardEuler import *
from src.Utils.RK2 import *
from src.Utils.Lambda2 import *
from src.Utils.Linear import *


class test_Constant1_Velocity_1D(unittest.TestCase):
    """
    Testing 1D problem with null velocity
    """
    def setUp(self):
        self.dim = 1
        self.boxLength = [1.]
        self.boxMin = [0.]
        self.nbElem = [20]
        self.timeStep = 1.
        self.FinalTime = 1.
        self.box = Box(dimension=self.dim,
                       length=self.boxLength,
                       minPosition=self.boxMin)
        self.velo = AnalyticalVariable(domain=self.box,
                                   dimension=self.dim, formula=self._velocity)
        self.scal = AnalyticalVariable(domain=self.box,
                                   dimension=1, formula=self._scalar)
        self.advec = Advection(self.velo, self.scal)
        self.cTspPb = ContinuousTransportProblem(advection=self.advec)
        self.cTspPb.discretize(dSpec=[self.nbElem])

    def testSolvingEuler(self):
        """ Testing Euler solution."""
        InterpolationMethod = Linear(grid=self.box.discreteDomain, gvalues=self.velo.discreteVariable)
        RemeshingMethod = Lambda2(grid=self.box.discreteDomain, gvalues=self.scal.discreteVariable)
        ODESolver = ForwardEuler(InterpolationMethod, self.box.discreteDomain.applyConditions)
        solver = ParticularSolver(ODESolver, InterpolationMethod, RemeshingMethod)
        self.cTspPb.setSolver(solver)
        self.cTspPb.solve(T=self.FinalTime, dt=self.timeStep)
        for i, p in self.box.discreteDomain.explore():
            self.assertAlmostEqual(self.scal.discreteVariable.values[i], self._scalar(p))
            [self.assertAlmostEqual(self.velo.discreteVariable.values[i][ii], self._velocity(p)[ii]) for ii in xrange(self.dim)]

    def testSolvingRK2(self):
        """ Testing RK2 solution."""
        InterpolationMethod = Linear(grid=self.box.discreteDomain, gvalues=self.velo.discreteVariable)
        RemeshingMethod = Lambda2(grid=self.box.discreteDomain, gvalues=self.scal.discreteVariable)
        ODESolver = RK2(InterpolationMethod, self.box.discreteDomain.applyConditions)
        solver = ParticularSolver(ODESolver, InterpolationMethod, RemeshingMethod)
        self.cTspPb.setSolver(solver)
        self.cTspPb.solve(T=self.FinalTime, dt=self.timeStep)
        for i, p in self.box.discreteDomain.explore():
            self.assertAlmostEqual(self.scal.discreteVariable.values[i], self._scalar(p))
            [self.assertAlmostEqual(self.velo.discreteVariable.values[i][ii], self._velocity(p)[ii]) for ii in xrange(self.dim)]

    def _velocity(self, x):
        return [1.]

    def _scalar(self, x):
        return x[0] * 10.


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(test_Constant1_Velocity_1D))
    return suite

if __name__ == "__main__":
    unittest.TextTestRunner(verbosity=2).run(suite())
