# -*- coding: utf-8 -*-
import unittest

from src.Problem.ContinuousTransportProblem import *
from src.Domain.Box import *
from src.Variable.AnalyticalVariable import *
from src.Operator.Advection import *
from src.Utils.ParticularSolver import *
from src.Utils.ForwardEuler import *
from src.Utils.RK2 import *
from src.Utils.Lambda2 import *
from src.Utils.Linear import *
from src.Utils.Printer import *


class test_Constant_Velocity_3D(unittest.TestCase):
    """
    Testing 3D problem with different constant velocities
    """
    def setUp(self):
        self.e = 0.55  # Accepted error between result and analytical result
        self.dim = 3
        self.boxLength = [2., 2., 2.]
        self.boxMin = [-1., -1., -1.]
        self.nbElem = [16, 16, 16]
        self.timeStep = 1.
        self.box = Box(dimension=self.dim,
                       length=self.boxLength,
                       minPosition=self.boxMin)

    def solveAndTest(self):
        self.velo = AnalyticalVariable(domain=self.box,
                                   dimension=self.dim, formula=self._velocity)
        self.scal = AnalyticalVariable(domain=self.box,
                                   dimension=1, formula=self._scalarInit, scalar=True)
        self.advec = Advection(self.velo, self.scal)
        self.cTspPb = ContinuousTransportProblem(advection=self.advec)
        self.cTspPb.discretize(dSpec=[self.nbElem])
        InterpolationMethod = Linear(grid=self.box.discreteDomain, gvalues=self.velo.discreteVariable)
        RemeshingMethod = Lambda2(grid=self.box.discreteDomain, gvalues=self.scal.discreteVariable)
        ODESolver = RK2(InterpolationMethod, self.box.discreteDomain.applyConditions)
        solver = ParticularSolver(ODESolver, InterpolationMethod, RemeshingMethod)
        self.cTspPb.setSolver(solver)
        #self.cTspPb.setPrinter(Printer(frequency=1, outputPrefix='res3D/RK2_', problem=self.cTspPb))
        self.FinalTime = self.boxLength[0] / self.v
        self.cTspPb.solve(T=self.FinalTime, dt=self.timeStep)
        for i in self.box.discreteDomain.explore():
            if abs(self._scalarInit(self.box.discreteDomain.points[i]) - self.scal.discreteVariable.values[i]) < self.e:
                self.assertTrue(abs(self._scalarInit(self.box.discreteDomain.points[i]) - self.scal.discreteVariable.values[i]) < self.e)
            else:
                print abs(self._scalarInit(self.box.discreteDomain.points[i]) - self.scal.discreteVariable.values[i])
            [self.assertAlmostEqual(self.velo.discreteVariable.values[i][ii], self._velocity(self.box.discreteDomain.points[i])[ii]) for ii in xrange(self.dim)]

    def testSolvingRK2_1(self):
        """Testing an advection around the box, v = (1,1,1) * 1/3 * dx"""
        self.v = (1. / 3.) * (self.boxLength[0] / (1. * self.nbElem[0]))
        self._velocity = lambda x: [self.v, -self.v, -self.v]
        self.solveAndTest()

    def testSolvingRK2_2(self):
        """Testing an advection around the box, v = (1,1,1) * 2/3 * dx"""
        self.v = (2. / 3.) * (self.boxLength[0] / (1. * self.nbElem[0]))
        self._velocity = lambda x: [self.v, -self.v, -self.v]
        self.solveAndTest()

    def testSolvingRK2_3(self):
        """Testing an advection around the box, v = (1,1,1) * 1/2 * dx"""
        self.v = (1. / 2.) * (self.boxLength[0] / (1. * self.nbElem[0]))
        self._velocity = lambda x: [self.v, -self.v, -self.v]
        self.solveAndTest()

    def testSolvingRK2_4(self):
        """Testing an advection around the box, v = (1,1,1) * 10/3 * dx"""
        self.v = (10. / 3.) * (self.boxLength[0] / (1. * self.nbElem[0]))
        self._velocity = lambda x: [self.v, -self.v, -self.v]
        self.solveAndTest()

    def testSolvingRK2_5(self):
        """Testing an advection around the box, v = (1,-1,-1) * 1/3 * dx"""
        self.v = (1. / 3.) * (self.boxLength[0] / (1. * self.nbElem[0]))
        self._velocity = lambda x: [self.v, -self.v, -self.v]
        self.solveAndTest()

    def testSolvingRK2_6(self):
        """Testing an advection around the box, v = (1,-1,-1) * 2/3 * dx"""
        self.v = (2. / 3.) * (self.boxLength[0] / (1. * self.nbElem[0]))
        self._velocity = lambda x: [self.v, -self.v, -self.v]
        self.solveAndTest()

    def testSolvingRK2_7(self):
        """Testing an advection around the box, v = (1,-1,-1) * 1/2 * dx"""
        self.v = (1. / 2.) * (self.boxLength[0] / (1. * self.nbElem[0]))
        self._velocity = lambda x: [self.v, -self.v, -self.v]
        self.solveAndTest()

    def testSolvingRK2_8(self):
        """Testing an advection around the box, v = (1,-1,-1) * 10/3 * dx"""
        self.v = (10. / 3.) * (self.boxLength[0] / (1. * self.nbElem[0]))
        self._velocity = lambda x: [self.v, -self.v, -self.v]
        self.solveAndTest()

    def _scalarInit(self, x):
        rr = (x[0] - 0.25) * (x[0] - 0.25) + (x[1] - 0.25) * (x[1] - 0.25) + (x[2] - 0.25) * (x[2] - 0.25)
        if rr < 0.25:
            return 1. - rr / 0.25
        else:
            return 0.

    def runTest(self):
        self.setUp()
        self.testSolvingRK2_1()


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(test_Constant_Velocity_3D))
    return suite

if __name__ == "__main__":
    unittest.TextTestRunner(verbosity=2).run(suite())
    #t = test_Constant_Velocity_3D()
    #t.setUp()
    #t.testSolvingRK2_1()
