# -*- coding: utf-8 -*-
import unittest

from src.Problem.ContinuousTransportProblem import *
from src.Domain.Box import *
from src.Variable.AnalyticalVariable import *
from src.Operator.Advection import *
from src.Utils.ParticularSolver import *
from src.Utils.ForwardEuler import *
from src.Utils.RK2 import *
from src.Utils.M6Prime import *
from src.Utils.Linear import *
from src.Utils.Printer import *
import math


class test_Rotating_Velocity_2D(unittest.TestCase):
    """
    Testing 2D problem with differents constant velocities
    """
    def setUp(self):
        self.e = 0.2  # Accepted error between result and analytical result
        self.dim = 2
        self.boxLength = [2., 2.]
        self.boxMin = [-1., -1.]
        self.nbElem = [200, 200]
        self.timeStep = 0.02
        self.box = Box(dimension=self.dim,
                       length=self.boxLength,
                       minPosition=self.boxMin)

    def testSolvingRK2_1(self):
        """Testing an advection around the box, v = (1,1) * 1/3 * dx"""
        self._velocity = lambda x: [- x[1] * 2 * math.acos(-1), x[0] * 2 * math.acos(-1)]
        self.velo = AnalyticalVariable(domain=self.box,
                                   dimension=self.dim, formula=self._velocity)
        self.scal = AnalyticalVariable(domain=self.box,
                                   dimension=1, formula=self._scalarInit, scalar=True)
        self.advec = Advection(self.velo, self.scal)
        self.cTspPb = ContinuousTransportProblem(advection=self.advec)
        self.cTspPb.discretize(dSpec=[self.nbElem])
        InterpolationMethod = Linear(grid=self.box.discreteDomain, gvalues=self.velo.discreteVariable)
        RemeshingMethod = M6Prime(grid=self.box.discreteDomain, gvalues=self.scal.discreteVariable)
        ODESolver = RK2(InterpolationMethod, self.box.discreteDomain.applyConditions)
        solver = ParticularSolver(ODESolver, InterpolationMethod, RemeshingMethod)
        self.cTspPb.setSolver(solver)
        self.cTspPb.setPrinter(Printer(frequency=1, outputPrefix='res2D/RK2_Rotate_JB_', problem=self.cTspPb))
        self.FinalTime = 1.
        self.cTspPb.solve(T=self.FinalTime, dt=self.timeStep)
        maxVal = 0.
        for i in self.box.discreteDomain.explore():
            if abs(self._scalarInit(self.box.discreteDomain.points[i]) - self.scal.discreteVariable.values[i]) > maxVal:
                maxVal = abs(self._scalarInit(self.box.discreteDomain.points[i]) - self.scal.discreteVariable.values[i])
        self.assertLess(maxVal, self.e)
        print "Max error : ", maxVal

    def _scalarInit(self, x):
        rr = math.sqrt(x[0] * x[0] + (x[1] - 0.5) * (x[1] - 0.5))
        if rr < 0.25:
            return (1. - rr * rr / 0.0625) ** 4
        else:
            return 0.

    def runTest(self):
        self.setUp()
        self.testSolvingRK2_1()


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(test_Rotating_Velocity_2D))
    return suite

if __name__ == "__main__":
    #unittest.TextTestRunner(verbosity=2).run(suite())
    t = test_Rotating_Velocity_2D()
    t.setUp()
    t.testSolvingRK2_1()
