# -*- coding: utf-8 -*-
import unittest

from src.Domain.Box import *
from src.Variable.AnalyticalVariable import *


class testCartesianGrid(unittest.TestCase):
    """
    DiscreteDomain test class
    """
    def setUp(self):
        self.nbElem = 4
        self.b1D = Box(dimension=1, length=[1.], minPosition=[0.])
        self.b1D.discretize([self.nbElem])
        self.b2D = Box(dimension=2, length=[1., 1.], minPosition=[0., 0.])
        self.b2D.discretize([self.nbElem, self.nbElem])
        self.b3D = Box(dimension=3, length=[1., 1., 1.], minPosition=[0., 0., 0.])
        self.b3D.discretize([self.nbElem, self.nbElem, self.nbElem])

    def testDimensions(self):
        """--> Testing element number for each CartesianGrid types"""
        self.assertEqual(self.b1D.discreteDomain.len(), self.nbElem)
        self.assertEqual(self.b2D.discreteDomain.len(), self.nbElem ** 2)
        self.assertEqual(self.b3D.discreteDomain.len(), self.nbElem ** 3)

    def testPoints(self):
        """--> Testing created Points positions and index"""
        d = 1. / (self.nbElem)
        for i in self.b1D.discreteDomain.explore():
            self.assertAlmostEqual(self.b1D.discreteDomain[i][0], 0. + i[0] * d)
        for i in self.b2D.discreteDomain.explore():
            self.assertAlmostEqual(self.b2D.discreteDomain[i][0], 0. + i[0] * d)
            self.assertAlmostEqual(self.b2D.discreteDomain[i][1], 0. + i[1] * d)
        for i in self.b3D.discreteDomain.explore():
            self.assertAlmostEqual(self.b3D.discreteDomain[i][0], 0. + i[0] * d)
            self.assertAlmostEqual(self.b3D.discreteDomain[i][1], 0. + i[1] * d)
            self.assertAlmostEqual(self.b3D.discreteDomain[i][2], 0. + i[2] * d)

    def testPointsValues(self):
        """--> Testing Points initialization (velocity and scalar)"""
        scalar = AnalyticalVariable(domain=self.b3D, dimension=1, formula=self.formula31D)
        velocity3D = AnalyticalVariable(domain=self.b3D, dimension=3, formula=self.formula3D)
        scalar.discretize()
        velocity3D.discretize()
        for i in self.b3D.discreteDomain.explore():
            self.assertAlmostEqual(scalar.discreteVariable.values[i], i[0] * 0.25 * i[1] * 0.25 * i[2] * 0.25)
            self.assertAlmostEqual(velocity3D.discreteVariable.values[i][0], (i[1] - i[0]) * 0.25)
            self.assertAlmostEqual(velocity3D.discreteVariable.values[i][1], -i[0] * 0.25 * i[2] * 0.25 - i[1] * 0.25)
            self.assertAlmostEqual(velocity3D.discreteVariable.values[i][2], i[0] * 0.25 * i[1] * 0.25 - i[2] * 0.25)

    def testPeriodicConditions(self):
        """--> Testing periodic conditions of positions"""
        for i in range(3):
            self.assertEqual(self.b3D.discreteDomain.applyConditions([0.5, 0.5, 0.5])[i], 0.5)
            self.assertEqual(self.b3D.discreteDomain.applyConditions([-0.5, 0.5, 0.5])[i], 0.5)
            self.assertEqual(self.b3D.discreteDomain.applyConditions([0.5, -0.5, 0.5])[i], 0.5)
            self.assertEqual(self.b3D.discreteDomain.applyConditions([0.5, 0.5, -0.5])[i], 0.5)
            self.assertEqual(self.b3D.discreteDomain.applyConditions([-0.5, -0.5, -0.5])[i], 0.5)
            self.assertEqual(self.b3D.discreteDomain.applyConditions([0.5, 0.5, 1.5])[i], 0.5)
            self.assertEqual(self.b3D.discreteDomain.applyConditions([0.5, 1.5, 0.5])[i], 0.5)
            self.assertEqual(self.b3D.discreteDomain.applyConditions([1.5, 0.5, 0.5])[i], 0.5)
            self.assertEqual(self.b3D.discreteDomain.applyConditions([1.5, 1.5, 1.5])[i], 0.5)

    def formula1D(self, position):
        """Fonction test 1D : f(x) = x**2  """
        return position[0] * position[0]

    def formula31D(self, position):
        """Fonction test 1D : f(x,y,z) = x*y*z  """
        return position[0] * position[1] * position[2]

    def formula2D(self, position):
        """Fonction test 2D : f(x,y) = (x - y,x + y) """
        return [position[0] - position[1], position[0] + position[1]]

    def formula3D(self, position):
        """Fonction test 3D : f(x,y,z) = (y - x,-xz - y,xy - z) """
        return [position[1] - position[0], -position[0] * position[2] - position[1], position[0] * position[1] - position[2]]


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(testCartesianGrid))
    return suite

if __name__ == "__main__":
    unittest.TextTestRunner(verbosity=2).run(suite())
