# -*- coding: utf-8 -*-
import unittest

from src.Problem.ContinuousTransportProblem import *
from src.Domain.Box import *
from src.Variable.AnalyticalVariable import *
from src.Operator.Advection import *
from src.Utils.RK2 import *
from src.Utils.Lambda1 import *
from src.Utils.Linear import *
from src.Utils.ParticularSolver import *


class testAdvectionDOp(unittest.TestCase):
    """
    AdvectionDOp test class
    """
    def setUp(self):
        self.dim = [1, 2, 3]
        self.nbElem = {}
        self.box = {}
        self.velo = {}
        self.scal = {}
        self.advec = {}
        self.InterpolationMethod = {}
        self.ODESolver = {}
        self.parts = {}
        self.ppos = {}
        self.pscal = {}
        self.advection = {}
        for dim in self.dim:
            self.nbElem[dim] = [10 for i in xrange(dim)]
            self.box[dim] = Box(dimension=dim, length=[1. for i in xrange(dim)], minPosition=[0. for i in xrange(dim)])
            self.velo[dim] = AnalyticalVariable(domain=self.box[dim], dimension=dim, formula=self._vitesse)
            self.scal[dim] = AnalyticalVariable(domain=self.box[dim], dimension=1, formula=self._scalaire, scalar=True)
            self.box[dim].discretize(spec=self.nbElem[dim])
            self.velo[dim].discretize()
            self.scal[dim].discretize()
            self.advec[dim] = Advection(velocity=self.velo[dim], scalar=self.scal[dim])
            self.InterpolationMethod[dim] = Linear(grid=self.box[dim].discreteDomain, gvalues=self.velo[dim].discreteVariable)
            self.ODESolver[dim] = RK2(self.InterpolationMethod[dim], dim=dim)
            self.parts[dim] = ParticleField(self.box[dim].discreteDomain)
            self.ppos[dim] = DiscreteVariable(domain=self.parts[dim], dimension=dim)
            self.parts[dim].setPositions(self.ppos[dim])
            self.pscal[dim] = DiscreteVariable(domain=self.parts[dim], dimension=1, initialValue=lambda x: 0., scalar=True)
            self.advection[dim] = AdvectionDOp(self.advec[dim])
            self.advection[dim].setResultVariable(respos=self.ppos[dim], resscal=self.pscal[dim])
            self.advection[dim].setMethod(self.ODESolver[dim])

    def testApplyOperator(self):
        """--> Testing Apply Advection Operator """
        for dim in self.dim:
            for d in xrange(dim):
                self.advection[dim].applyOperator(t=0., dt=0.05, splittingDirection=d)
                for i in self.box[dim].discreteDomain.explore():
                    point = [self.box[dim].discreteDomain.points[i][dd] + 0.05 if dd == d else self.box[dim].discreteDomain.points[i][dd] for dd in xrange(dim)]
                    self.assertAlmostEqual(self.ppos[dim].values[i][d], self.box[dim].discreteDomain.applyConditions(point)[d])
                    self.assertAlmostEqual(self.pscal[dim].values[i], self.scal[dim].discreteVariable.values[i])
            for d in xrange(dim):
                self.advection[dim].applyOperator(t=0., dt=0.15, splittingDirection=d)
                for i in self.box[dim].discreteDomain.explore():
                    point = [self.box[dim].discreteDomain.points[i][dd] + 0.15 if dd == d else self.box[dim].discreteDomain.points[i][dd] for dd in xrange(dim)]
                    self.assertAlmostEqual(self.ppos[dim].values[i][d], self.box[dim].discreteDomain.applyConditions(point)[d])
                    self.assertAlmostEqual(self.pscal[dim].values[i], self.scal[dim].discreteVariable.values[i])

    def _vitesse(self, x):
        return [1. for i in x]

    def _scalaire(self, x):
        if x[0] < 0.5:
            return 1.
        else:
            return 0.


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(testAdvectionDOp))
    return suite

if __name__ == "__main__":
    unittest.TextTestRunner(verbosity=2).run(suite())
