# -*- coding: utf-8 -*-
import unittest

from src.Domain.Box import *
from src.Variable.AnalyticalVariable import *
from src.Variable.DiscreteVariable import *
from src.Domain.ParticleField import *
from src.Operator.RemeshingDOp import *
from src.Utils.Lambda2 import *


class testRemeshDOp(unittest.TestCase):
    """
    RemeshDOp test class
    """
    def setUp(self):
        self.nbElem = 10
        self.box = Box(dimension=1, length=[1.], minPosition=[0.])
        self.box.discretize(spec=[self.nbElem])
        self.scal = AnalyticalVariable(dimension=1, domain=self.box, formula=lambda x: 0., scalar=True)
        self.scal.discretize()
        self.parts = ParticleField(self.box.discreteDomain)
        self.ppos = DiscreteVariable(domain=self.parts, dimension=1)
        self.parts.setPositions(self.ppos)
        self.pscal = DiscreteVariable(domain=self.parts, dimension=1, scalar=True)
        self.pbloc = DiscreteVariable(domain=self.parts, dimension=1, scalar=True, integer=True)
        self.ptag = DiscreteVariable(domain=self.parts, dimension=1, scalar=True, integer=True)
        self.remesh = RemeshingDOp(partPositions=self.ppos,
                                   partScalar=self.pscal,
                                   partBloc=self.pbloc,
                                   partTag=self.ptag,
                                   resscal=self.scal.discreteVariable,
                                   method=Lambda2(grid=self.box.discreteDomain, gvalues=self.scal.discreteVariable))

    def testApplyOperator_noTag(self):
        """--> Testing Apply Remesh Operator without tag"""
        cfl = 1.
        dt = cfl * self.box.discreteDomain.elementSize[0]
        # Initialisation des positions et scalaires
        # Pas de tag. Toutes les particules appartiennent à des blocs Gauche (0)
        for i in self.parts.explore():
            self.pscal.values[i] = 1.
            self.pbloc.values[i] = 0
            self.ptag.values[i] = 0
            self.ppos.values[i] = [i[0] * self.box.discreteDomain.elementSize[0] + dt * 0.1]
        for i in self.box.discreteDomain.explore():
            self.scal.discreteVariable.values[i] = 0.
        self.remesh.applyOperator(t=0., dt=dt, splittingDirection=0)
        [self.assertAlmostEqual(self.scal.discreteVariable.values[i], 1.) for i in self.box.discreteDomain.explore()]
        # Pas de tag. Toutes les particules appartiennent à des blocs Gauche (0)
        for i in self.parts.explore():
            self.pscal.values[i] = 1.
            self.pbloc.values[i] = 0
            self.ptag.values[i] = 0
            self.ppos.values[i] = [i[0] * self.box.discreteDomain.elementSize[0] - dt * 0.9]
        for i in self.box.discreteDomain.explore():
            self.scal.discreteVariable.values[i] = 0.
        self.remesh.applyOperator(t=0., dt=dt, splittingDirection=0)
        [self.assertAlmostEqual(self.scal.discreteVariable.values[i], 1.) for i in self.box.discreteDomain.explore()]
        # Pas de tag. Toutes les particules appartiennent à des blocs Centrés (1)
        for i in self.parts.explore():
            self.pscal.values[i] = 1.
            self.pbloc.values[i] = 1
            self.ptag.values[i] = 0
            self.ppos.values[i] = [i[0] * self.box.discreteDomain.elementSize[0] + dt * 0.9]
        for i in self.box.discreteDomain.explore():
            self.scal.discreteVariable.values[i] = 0.
        self.remesh.applyOperator(t=0., dt=dt, splittingDirection=0)
        [self.assertAlmostEqual(self.scal.discreteVariable.values[i], 1.) for i in self.box.discreteDomain.explore()]
        # Pas de tag. Toutes les particules appartiennent à des blocs Centrés (1)
        for i in self.parts.explore():
            self.pscal.values[i] = 1.
            self.pbloc.values[i] = 1
            self.ptag.values[i] = 0
            self.ppos.values[i] = [i[0] * self.box.discreteDomain.elementSize[0] - dt * 0.1]
        for i in self.box.discreteDomain.explore():
            self.scal.discreteVariable.values[i] = 0.
        self.remesh.applyOperator(t=0., dt=dt, splittingDirection=0)
        [self.assertAlmostEqual(self.scal.discreteVariable.values[i], 1.) for i in self.box.discreteDomain.explore()]

    def testApplyOperator_noTag_shifted(self):
        """--> Testing Apply Remesh Operator without tag, shifted velocity"""
        cfl = 1.
        dt = cfl * self.box.discreteDomain.elementSize[0]
        # Initialisation des positions et scalaires
        # Pas de tag. Toutes les particules appartiennent à des blocs Gauche (0)
        for i in self.parts.explore():
            self.pscal.values[i] = 1.
            self.pbloc.values[i] = 0
            self.ptag.values[i] = 0
            self.ppos.values[i] = [i[0] * self.box.discreteDomain.elementSize[0] + dt * 5.1]
        for i in self.box.discreteDomain.explore():
            self.scal.discreteVariable.values[i] = 0.
        self.remesh.applyOperator(t=0., dt=dt, splittingDirection=0)
        [self.assertAlmostEqual(self.scal.discreteVariable.values[i], 1.) for i in self.box.discreteDomain.explore()]
        # Pas de tag. Toutes les particules appartiennent à des blocs Gauche (0)
        for i in self.parts.explore():
            self.pscal.values[i] = 1.
            self.pbloc.values[i] = 0
            self.ptag.values[i] = 0
            self.ppos.values[i] = [i[0] * self.box.discreteDomain.elementSize[0] - dt * 5.9]
        for i in self.box.discreteDomain.explore():
            self.scal.discreteVariable.values[i] = 0.
        self.remesh.applyOperator(t=0., dt=dt, splittingDirection=0)
        [self.assertAlmostEqual(self.scal.discreteVariable.values[i], 1.) for i in self.box.discreteDomain.explore()]
        # Pas de tag. Toutes les particules appartiennent à des blocs Centrés (1)
        for i in self.parts.explore():
            self.pscal.values[i] = 1.
            self.pbloc.values[i] = 1
            self.ptag.values[i] = 0
            self.ppos.values[i] = [i[0] * self.box.discreteDomain.elementSize[0] + dt * 5.9]
        for i in self.box.discreteDomain.explore():
            self.scal.discreteVariable.values[i] = 0.
        self.remesh.applyOperator(t=0., dt=dt, splittingDirection=0)
        [self.assertAlmostEqual(self.scal.discreteVariable.values[i], 1.) for i in self.box.discreteDomain.explore()]
        # Pas de tag. Toutes les particules appartiennent à des blocs Centrés (1)
        for i in self.parts.explore():
            self.pscal.values[i] = 1.
            self.pbloc.values[i] = 1
            self.ptag.values[i] = 0
            self.ppos.values[i] = [i[0] * self.box.discreteDomain.elementSize[0] - dt * 5.1]
        for i in self.box.discreteDomain.explore():
            self.scal.discreteVariable.values[i] = 0.
        self.remesh.applyOperator(t=0., dt=dt, splittingDirection=0)
        [self.assertAlmostEqual(self.scal.discreteVariable.values[i], 1.) for i in self.box.discreteDomain.explore()]


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(testRemeshDOp))
    return suite

if __name__ == "__main__":
    unittest.TextTestRunner(verbosity=2).run(suite())
